#include "lw_fftw3.h"


/****************************************************************************/
/* FFT Convolution by parts using the fftw3 library */

/* filter should be an YSIG, M = filter->size should be < in->size,
   filter->x0 and filter->dx should be such that i0 = -x0/dx is an integer
   pointing to the X=0 index of the filter. 
   0 <= i0 < M (hence x0 should be < 0)
   m_pos = M-1-i0 is the filter size with X > 0
   m_neg = i0 is the filter size with X < 0
   M = m_pos+m_neg+1

   size_window should be > M+1
*/
  
static void lwfftw_get_filter_info(SIGNAL filter, int *p_m_pos, int *p_m_neg) { 
  if ( filter->type != YSIG ) {Errorf("lwfftw_get_filter_info(): Filter should be an YSIG !!");}

  *p_m_neg = (int) (0.5-filter->x0/filter->dx);
  *p_m_pos = filter->size-1-*p_m_neg;

# ifdef LWFFTW_DEBUG 
  Printf("Filter size: %d\tm+: %d\tm-: %d\n",filter->size,*p_m_pos,*p_m_neg);
# endif

  /* Some consistency checking */
  if ( (*p_m_neg < 0 ) || (*p_m_neg >= filter->size) ) {
    Errorf("lwfftw_get_filter_info(): i0 (=%d) should be in range [0,%d[ !!",
	   *p_m_neg,filter->size);
  }
}


/* Computing the rounded UP binary logarithm for a 32 bit integer. argument should be > 1 */
static int lwfftw_ceil_log2(int n) {
  int res = 1;
  int pow2res = 2;

  while(pow2res < n) {
    pow2res *= 2;
    res += 1;
  }
  
  return res;
}

/* Computing optimal widow_size for convolution */
static int lwfft_optimal_window_size(SIGNAL in, SIGNAL filter, int *p_num_windows) {
  int min_log2_window_size,max_log2_window_size,log2_window_size,window_size,num_windows;
  int L,optimal_window_size;
  double complexity,min_complexity; /* double to avoid overflow problems */

  /* min_window_size is the smallest power of two such that min_window_size >= filter->size */
  min_log2_window_size = lwfftw_ceil_log2(filter->size);
  /* max_window_size is the smallest power of two such that max_window_size >= in->size+filter->size-1 */
  max_log2_window_size = lwfftw_ceil_log2(in->size+filter->size-1);

# ifdef LWFFTW_DEBUG 
  Printf("Filter size: %d\tlog2: %d\nMax tot: %d\tlog2: %d\n",
	 filter->size,min_log2_window_size,in->size+filter->size-1,max_log2_window_size);
# endif

  min_complexity = 0;
  for(log2_window_size = min_log2_window_size; log2_window_size <= max_log2_window_size; log2_window_size += 1) {
    window_size = 1<<log2_window_size;
    L = window_size-filter->size+1; /* Number of points NOT affected by border effects for each FFT on window_size */
    /* num_windows is the smallest integer such that num_windows >= in->size/L */
    num_windows = in->size/L + ((in->size%L == 0) ? 0 : 1);
    /* Convolution involve num_windows FFTs for in FT, one FFT for filter FT and num_windows iFFT */
    complexity = (2.0*num_windows+1.0)*window_size*log2_window_size;

#  ifdef LWFFTW_DEBUG 
    Printf("window_size: %d\tnum_windows: %d\tcomplexity: %.1f\tcomplexity: %.1f\n",
	   window_size,num_windows,complexity,1.0*num_windows*window_size*log2_window_size);
#  endif

    if( (min_complexity == 0) || (complexity <= min_complexity) ) {
      min_complexity = complexity;
      optimal_window_size = window_size;
      *p_num_windows = num_windows;
    }
  }

  return optimal_window_size;
}

/* Padding <input> for fft convolution with <filter> using windows  of size <window_size> */
static int lwfftw_init_with_border(SIGNAL in, SIGNAL filter, int window_size, SIGNAL in_padded, int border_type) {
  int i,j,m_pos,m_neg;
  int L; /* number of coeff with no border effect for each part of the fft convol */
  int num_windows, padded_size;

  lwfftw_get_filter_info(filter, &m_pos, &m_neg);
  L = window_size-m_pos-m_neg;
  if(L <= 0) {
    Errorf("lwfftw_init_with_border(): window_size (=%d) should be larger than %d",
	   window_size, filter->size);
  }

  num_windows = in->size/L + ((in->size%L == 0) ? 0 : 1); 
  padded_size = (num_windows-1)*L+window_size;

  SizeSignal(in_padded, padded_size, YSIG);
  in_padded->dx = in->dx;
  in_padded->x0 = -m_pos;

  /* Middle */
  for(i=m_pos; i < in->size+m_pos; i++) {
    in_padded->Y[i] = in->Y[i-m_pos];
  }
  
  /* Left */
  switch( border_type) {
  case BorderPad :
    for(i=0; i<m_pos; i++) {in_padded->Y[i] = in->Y[0];}
    break;
    
  case BorderPad0 :
    for(i=0; i<m_pos; i++) {in_padded->Y[i] = 0.0;}
    break;
    
  case BorderPeriodic :
    for(i=0; i<m_pos; i++) {
      j = (i-m_pos);
      /* Complication due to unclear def of % in C for negative numerator */
      while(j<0) {j+=in->size;}
      j = j%in->size;
      in_padded->Y[i] =  in->Y[j];
    }
    break;
    
  case BorderMirror :
    for(i=0; i<m_pos; i++) {
      j = (i-m_pos);
      /* Complication due to unclear def of % in C for negative numerator */
      while(j<0) {j += (2*in->size);}
      j = j%(2*in->size);
      if(j>=in->size) {j = 2*in->size-1-j;}
      in_padded->Y[i] = in->Y[j];
    }
    break;
    
  default:
    Errorf("lwfftw_init_with_border() : Bad Border type !");
  }
  
  /* Right */
  switch( border_type ) {
  case BorderPad :
    for(i=in->size+m_pos; i<in_padded->size; i++) {in_padded->Y[i] = in->Y[in->size-1];}
    break;
    
  case BorderPad0 :
    for(i=in->size+m_pos; i<in_padded->size; i++) {in_padded->Y[i] = 0.0;}
    break;
    
  case BorderPeriodic :
    for(i=in->size+m_pos; i<in_padded->size; i++) {
      j = (i-m_pos)%in->size;
      in_padded->Y[i] = in->Y[j];
    }
    break;
    
  case BorderMirror :
    for(i=in->size+m_pos; i<in_padded->size; i++) {
      j = (i-m_pos)%(2*in->size);
      if(j>=in->size) {j = 2*in->size-1-j;}
      in_padded->Y[i] = in->Y[j];
    }    
    break;
    
  default:
    Errorf("lwfftw_init_with_border() : Bad Border type !");
  }

  return num_windows;
}

/* Forward FT of in_padded */
static int lwfftw_multiple_r2c_1d(SIGNAL in_padded, int filter_logical_size, int window_size, int num_windows,
				  SIGNAL ft_in_padded, unsigned fftw_flags) {
  lw_fftw_plan p;
  int window_step = window_size-(filter_logical_size-1);
  int window_size_fft = window_size/2+1;

  /* Consistency checking */
  if( num_windows < 1 ) {Errorf("lwfftw_multiple_r2c_1d(): num_windows should be >= 1 !!");}
  if( window_step < 1 ) {
    Errorf("lwfftw_multiple_r2c_1d(): window_step (=window_size-(filter_logical_size-1)) should be >= 1 !!");
  }
  if( in_padded->size < (num_windows-1)*window_step+window_size ) {
    Errorf("lwfftw_multiple_r2c_1d(): Padded signal size is too small !!");
  }

  /* out will contain the output complex results: out->size = 2*(in->size/2+1) */
  SizeSignal(ft_in_padded,num_windows*2*window_size_fft,YSIG);

  p = LW_FFTW_NAMES(plan_many_dft_r2c)(1, &window_size, num_windows,
				       in_padded->Y, NULL, 1, window_step,
				       (lw_fftw_complex *) ft_in_padded->Y, NULL, 1, window_size_fft,
				       fftw_flags);
  if (p == NULL) {Errorf("lwfftw_multiple_r2c_1d(): Problem while making plan !!");}

  LW_FFTW_NAMES(execute)(p);
  
  LW_FFTW_NAMES(destroy_plan)(p);

  return window_size_fft;
}

/* WARNING: this backward transform destroy input signal */
static void lwfftw_multiple_c2r_1d(SIGNAL in, SIGNAL out, int logical_size, int howmany, unsigned fftw_flags) {
  lw_fftw_plan p;
  int size_window_fft = logical_size/2+1;

  if( howmany*2*size_window_fft != in->size ) {
    Errorf("lwfftw_multiple_c2r_1d(): Input size (=%d) is not compatible with logical_size (=%d) "
	   "and howmany (=%d) !!",in->size, logical_size, howmany);
  }

  /* out will contain the output real results: out->size = howmany*logical_size */
  SizeSignal(out,howmany*logical_size,YSIG);

  p = LW_FFTW_NAMES(plan_many_dft_c2r)(1, &logical_size, howmany,
				       (lw_fftw_complex *) in->Y, NULL, 1, size_window_fft,
				       out->Y, NULL, 1, logical_size,
				       fftw_flags);

  if (p == NULL) {Errorf("lwfftw_multiple_c2r_1d(): Problem while making plan !!");}

  LW_FFTW_NAMES(execute)(p);
  
  LW_FFTW_NAMES(destroy_plan)(p);
}


void C_Lwfftw_convol(char **argv) {
  char *action;
 
  argv = ParseArgv(argv,tWORD,&action,-1);

  if (!strcmp(action,"window_size")) { /* 'window_size' action */
    int optimal_window_size, num_windows;
    SIGNAL in,filter;
    LISTV reslv;

    argv = ParseArgv(argv,tSIGNALI,&in,tSIGNALI,&filter,0);
    
    optimal_window_size = lwfft_optimal_window_size(in,filter,&num_windows);

    reslv = TNewListv();
    SetLengthListv(reslv,2);
    SetListvNthFloat(reslv,0,(LWFLOAT) optimal_window_size);
    SetListvNthFloat(reslv,1,(LWFLOAT) num_windows);
    SetResultValue(reslv);

 
  } else if (!strcmp(action,"pad")) { /* 'pad' action */
    char *borderName, opt;
    int window_size,border_type;
    SIGNAL in,filter,in_padded;
    
    argv = ParseArgv(argv,tSIGNALI,&in,tSIGNALI,&filter,tINT,&window_size,tSIGNAL,&in_padded,-1);

    border_type = BorderMirror;
    while(opt = ParseOption(&argv)) {
      switch(opt) {
      case 'b':
	argv = ParseArgv(argv,tSTR,&borderName,-1);
	if(!strcmp(borderName,"pad"))
	  border_type =  BorderPad;
	else if(!strcmp(borderName,"mir"))
	  border_type = BorderMirror;
	else if(!strcmp(borderName,"pad0"))
	  border_type = BorderPad0;
	else if(!strcmp(borderName,"per"))
	  border_type = BorderPeriodic;
	else 
	  Errorf("Undefined border effect: %s",borderName);
	break;
      default: 
	ErrorOption(opt);
      }
    }    
    NoMoreArgs(argv);

    SetResultInt(lwfftw_init_with_border(in, filter, window_size, in_padded, border_type));


  } else if (!strcmp(action,"filter_info")) { /* 'filter_info' action */
    int m_pos, m_neg;
    SIGNAL filter;
    LISTV reslv;

    argv = ParseArgv(argv,tSIGNALI,&filter,0);

    lwfftw_get_filter_info(filter, &m_pos, &m_neg);

    reslv = TNewListv();
    SetLengthListv(reslv,2);
    SetListvNthFloat(reslv,0,(LWFLOAT) m_neg);
    SetListvNthFloat(reslv,1,(LWFLOAT) m_pos);
    SetResultValue(reslv);


  } else if (!strcmp(action,"multi_r2c")) { /*** 'multi_r2c' action */
    int window_size, num_windows, filter_logical_size;
    SIGNAL in_padded, ft_in_padded;

    argv = ParseArgv(argv,tSIGNALI,&in_padded,tINT,&filter_logical_size,tINT,&window_size,tINT,&num_windows,
		     tSIGNAL,&ft_in_padded,0);

    SetResultInt(lwfftw_multiple_r2c_1d(in_padded, filter_logical_size, window_size, num_windows,
					ft_in_padded, FFTW_ESTIMATE));
      

  } else if (!strcmp(action,"perio_filter")) { /*** 'perio_filter' action */
    /* Construct filter perio*/
    int window_size, m_pos, m_neg, i, m_pos_alt, m_neg_alt;
    SIGNAL filter_re, filter_im, filter_perio;
    LISTV reslv;
    VALUE val; 
    char *type;

    argv = ParseArgv(argv,tSIGNALI,&filter_re,tVALOBJ,&val,tINT,&window_size,tSIGNAL,&filter_perio,0);
    type = GetTypeValue(val);
    if (type == signalType || type == signaliType) { 
      filter_im = (SIGNAL) ValueOf(val);
    } else if (type == nullType) {
      filter_im = NULL;
    } else {
      Errorf("Bad type '%s' for argument <filter_im>",type);
    }

    lwfftw_get_filter_info(filter_re, &m_pos, &m_neg);
    if (filter_im == NULL) {
      SizeSignal(filter_perio, window_size, YSIG);
    } else {
      lwfftw_get_filter_info(filter_im, &m_pos_alt, &m_neg_alt);
      if ( (m_pos != m_pos_alt) || (m_neg != m_neg_alt) ) {
	Errorf("Real and Imaginary parts of the filter are not compatible !!");
      }
      SizeSignal(filter_perio, 2*window_size, YSIG);
    }
    
    if (filter_im == NULL) {
      for(i=0;i <= m_pos;i++) {filter_perio->Y[i] = filter_re->Y[m_neg+i];}
      for(;i < window_size-m_neg;i++) {filter_perio->Y[i] = 0.0;}
      for(;i < window_size;i++) {filter_perio->Y[i] = filter_re->Y[i-(window_size-m_neg)];}
    } else {
      for(i=0;i <= m_pos;i++) {
	filter_perio->Y[2*i] = filter_re->Y[m_neg+i];
	filter_perio->Y[2*i+1] = filter_im->Y[m_neg+i];
      }
      for(;i < window_size-m_neg;i++) {
	filter_perio->Y[2*i] = 0.0;
	filter_perio->Y[2*i+1] = 0.0;
      }
      for(;i < window_size;i++) {
	filter_perio->Y[2*i] = filter_re->Y[i-(window_size-m_neg)];
	filter_perio->Y[2*i+1] = filter_im->Y[i-(window_size-m_neg)];
      }
    }

    reslv = TNewListv();
    SetLengthListv(reslv,2);
    SetListvNthFloat(reslv,0,(LWFLOAT) m_neg);
    SetListvNthFloat(reslv,1,(LWFLOAT) m_pos);
    SetResultValue(reslv);


  } else if (!strcmp(action,"complex_prod")) { /*** 'complex_prod' action */
    int num_windows,i,k,offset,num_complexes,ir,ii;
    SIGNAL ft_in, ft_filter, ft_prod;
    LWFLOAT in_re,in_im;
    
    argv = ParseArgv(argv,tSIGNALI,&ft_in,tSIGNALI,&ft_filter,tINT,&num_windows,tSIGNAL_,NULL,&ft_prod,0);

    if( ft_in->size != num_windows*ft_filter->size ) {Errorf("Expecting ft_in->size == num_windows*ft_filter->size");}
    num_complexes = ft_filter->size/2;
    /* Printf("num_complexes: %d\tft_filter->size: %d\n",num_complexes,ft_filter->size); */
    if( 2*num_complexes != ft_filter->size) {Errorf("ft_filter size should be even !!");}

    if(ft_prod == NULL) { /* IN PLACE transform */
      ft_prod = ft_in;
    } else {
      SizeSignal(ft_prod,ft_in->size,YSIG);
    }

    for(k =0; k<num_windows; k++) {
      offset = k*ft_filter->size;
      for(i=0; i<num_complexes;i++) {
	ir = 2*i;
	ii = 2*i+1;
	in_re = ft_in->Y[offset+ir];
	in_im = ft_in->Y[offset+ii];
	ft_prod->Y[offset+ir] = in_re*ft_filter->Y[ir]-in_im*ft_filter->Y[ii];
	ft_prod->Y[offset+ii] = in_re*ft_filter->Y[ii]+in_im*ft_filter->Y[ir];
      }
    }
    
  } else if (!strcmp(action,"recons_c2mp")) { /*** 'multi_c2r' action */
    int window_size, num_windows, first_good, size_good, window_step;
    int m_pos, m_neg, i, i_start, i_end, j, k, first_k, first_i_start_offset, last_k, last_i_end_offset, offset, i_complex;
    SIGNAL complex_in, filter, sig_module, sig_phase;
    LWFLOAT norm, re, im;

    argv = ParseArgv(argv,tSIGNALI,&complex_in,tSIGNALI,&filter,
		     tINT,&window_size,tINT,&num_windows,
		     tINT,&first_good,tINT,&size_good,
		     tSIGNAL,&sig_module,tSIGNAL_,NULL,&sig_phase,
		     tFLOAT_,1.0,&norm,0);

    lwfftw_get_filter_info(filter, &m_pos, &m_neg);
    if(first_good<m_pos) {Errorf("Expecting first_good >= m_pos");}
    window_step = window_size-(filter->size-1);
    if(window_step<1) {Errorf("Expecting window_step >= 1");}

    first_k = (first_good-m_pos)/window_step;
    first_i_start_offset = (first_good-m_pos)%window_step;

    last_k = (first_good+size_good-1-m_pos)/window_step;
    last_i_end_offset = (first_good+size_good-1-m_pos)%window_step;
    if(last_k >= num_windows) {Errorf("(first_good+size_good-1) is out of bound !!");}

    
    /* Reconstruct the convolution */
    SizeSignal(sig_module,size_good,YSIG);
    if( sig_phase != NULL ) {SizeSignal(sig_phase,size_good,YSIG);}
    j = 0;
    norm *= 1.0/window_size;
    for(k=first_k;  k<=last_k; k++) {
      i_start = m_pos;
      if(k==first_k) {i_start += first_i_start_offset;}

      i_end = window_size-m_neg-1;
      if(k==last_k) {i_end = m_pos+last_i_end_offset;}

      offset = k*window_size;

      for(i=i_start;i<=i_end;i++) {
	i_complex = 2*(offset+i);
	re = complex_in->Y[i_complex];
	im = complex_in->Y[i_complex+1];
	sig_module->Y[j] = norm*hypot(re,im);
	if( sig_phase != NULL ) {
	  sig_phase->Y[j] = atan2(im,re);
	}
	j++;
      }
    }
    sig_module->lastp = j-1;
    if( sig_phase != NULL ) {sig_phase->lastp = j-1;}


  } else if (!strcmp(action,"multi_c2r")) { /*** 'multi_c2r' action */
    int window_size, num_windows, first_good, size_good, window_step;
    int m_pos, m_neg, i, i_start, i_end, j, k, first_k, first_i_start_offset, last_k, last_i_end_offset, offset;
    SIGNAL ft_in, filter, out;
    LWFLOAT norm;

    argv = ParseArgv(argv,tSIGNALI,&ft_in,tSIGNALI,&filter,
		     tINT,&window_size,tINT,&num_windows,
		     tINT,&first_good,tINT,&size_good,tSIGNAL,&out,
		     tFLOAT_,1.0,&norm,0);

    lwfftw_get_filter_info(filter, &m_pos, &m_neg);
    if(first_good<m_pos) {Errorf("Expecting first_good >= m_pos");}
    window_step = window_size-(filter->size-1);
    if(window_step<1) {Errorf("Expecting window_step >= 1");}

    first_k = (first_good-m_pos)/window_step;
    first_i_start_offset = (first_good-m_pos)%window_step;

    last_k = (first_good+size_good-1-m_pos)/window_step;
    last_i_end_offset = (first_good+size_good-1-m_pos)%window_step;
    if(last_k >= num_windows) {Errorf("(first_good+size_good-1) is out of bound !!");}

    /* IFFT */
    lwfftw_multiple_c2r_1d(ft_in, out, window_size, num_windows, FFTW_ESTIMATE);
  
    /* Reconstruct the convolution IN PLACE */
    j = 0;
    norm *= 1.0/window_size;
    for(k=first_k;  k<=last_k; k++) {
      i_start = m_pos;
      if(k==first_k) {i_start += first_i_start_offset;}

      i_end = window_size-m_neg-1;
      if(k==last_k) {i_end = m_pos+last_i_end_offset;}

      offset = k*window_size;

      for(i=i_start;i<=i_end;i++) {
	out->Y[j] = norm*out->Y[offset+i];
	j++;
      }
    }
    out->size = size_good;
    out->lastp = j-1;


  } else {
    Errorf("Unknow action '%s'",action);
  }

}
