#if !defined(LW_FFTW_NO_THREAD)
# include <pthread.h>
int lwfftw_multhr_nthreads = 1;
# define LWFFTW_MAX_NUM_THREADS 32
static pthread_t lwfftw_multhr_threads[LWFFTW_MAX_NUM_THREADS];
static pthread_attr_t lwfftw_multhr_attr;
static pthread_mutex_t lwfftw_multhr_next_plan_mutex = PTHREAD_MUTEX_INITIALIZER;
#endif


#include "lw_fftw3.h"


/* ********************************************
   Utility function for Multithreaded functions
   ******************************************** */
static lw_fftw_plan *next_plan_plans;
static int next_plan_step;
static int next_plan_next_index;
static int next_plan_num_plan;

static lw_fftw_plan lwfftw_multhr_next_plan(void) {
  int next_index = next_plan_next_index;

  if(next_index < 0) { /* No more */
    return NULL;
  }

# if !defined(LW_FFTW_NO_THREAD)
  if (pthread_mutex_lock(&lwfftw_multhr_next_plan_mutex)) {Errorf("lwfftw_multhr_next_plan(): Error while locking mutex !!");}
# endif

  next_plan_next_index += next_plan_step;
  if(next_plan_next_index >= next_plan_num_plan) {
    next_plan_next_index = (next_plan_next_index%next_plan_step)+1; /* going thoought next phase of indexes */
    if(next_plan_next_index == next_plan_step) {
      next_plan_next_index = -1; /* no next */
    }
  }
  
# if !defined(LW_FFTW_NO_THREAD)
  if (pthread_mutex_unlock(&lwfftw_multhr_next_plan_mutex)) {Errorf("lwfftw_multhr_next_plan(): Error while unlocking mutex !!");}
# endif

  Printf("%d ",next_index);
  return next_plan_plans[next_index];
}

static void *lwfftw_multhr_thread_execute_func(void *dum) {
  lw_fftw_plan p;
  while (p = lwfftw_multhr_next_plan()) {
    LW_FFTW_NAMES(execute)(p); /* This is the only thread safe fftw3 function */
  }

  return NULL;
}

/* **************************************
   Multiple Complex to Complex Transforms
   ************************************** */
void lwfftw_multhr_multiple_c2c_1d(SIGNAL in, SIGNAL out,
				   int logical_size, int num_windows, int logical_step,
				   int fftw_sign, unsigned fftw_flags) {
  int i, num_threads;
  lw_fftw_plan *plans;

  /* input signal should contain at least (num_windows-1)*logical_step+logical_size complex numbers */
  if( in->size < 2*((num_windows-1)*logical_step+logical_size) ) {
    Errorf("lwfftw_multhr_multiple_c2c_1d(): Input signal size is too small !!");
  }

  /* out will contain the output complex results: out->size = num_windows*2*logical_size */
  SizeSignal(out, num_windows*2*logical_size, YSIG);

  /* Creating nun_windows plan */
  plans = (lw_fftw_plan *) TMalloc(num_windows*sizeof(lw_fftw_plan));
  for(i=0; i<num_windows; i++) {
    plans[i] =  LW_FFTW_NAMES(plan_dft_1d)( logical_size,
					    (lw_fftw_complex *) (in->Y+2*logical_step*i),
					    (lw_fftw_complex *) (out->Y+2*logical_size*i),
					    fftw_sign, fftw_flags);
    if (plans[i] == NULL) {Errorf("lwfftw_multhr_multiple_c2c_1d(): Problem while making plan no %d!!",i);}
  }
  


  /* initialise next_plan() */
  next_plan_plans = plans;
  next_plan_step = num_windows; /* i.e. nthreads = 1 */
  next_plan_next_index = 0;
  next_plan_num_plan = num_windows;

  /* Computing */
#if !defined(LW_FFTW_NO_THREAD)
  if(lwfftw_multhr_nthreads > LWFFTW_MAX_NUM_THREADS) {
    num_threads = LWFFTW_MAX_NUM_THREADS;
    Warningf("num_threads (%d) too large, using %d threads.\n",lwfftw_multhr_nthreads, LWFFTW_MAX_NUM_THREADS);
  } else {
    num_threads = lwfftw_multhr_nthreads;
  }
  if(num_threads <= 1) {/* no need for extra threads */
    lwfftw_multhr_thread_execute_func(NULL);
  } else {
    next_plan_step = num_windows/num_threads;
    /* Shutdown fftw3 internal multithread capability */
    LW_FFTW_NAMES(plan_with_nthreads)(1);
    /* Initialize and set thread detached attribute */
    pthread_attr_init(&lwfftw_multhr_attr);
    pthread_attr_setdetachstate(&lwfftw_multhr_attr, PTHREAD_CREATE_JOINABLE);
    /* starting the threads */
    for(i=0; i<num_threads; i++) { 
      if ( pthread_create(&lwfftw_multhr_threads[i], &lwfftw_multhr_attr, lwfftw_multhr_thread_execute_func, NULL) ) {
	Errorf("lwfftw_multhr_multiple_c2c_1d(): pthread_create error!!\n");
      }
    }
    /* Free attribute and wait for the other threads */
    pthread_attr_destroy(&lwfftw_multhr_attr);
    for(i=0; i<num_threads; i++) {
      if ( pthread_join(lwfftw_multhr_threads[i], NULL) ) {Errorf("lwfftw_multhr_multiple_c2c_1d(): pthread_join error!!\n");}
    }
    /* Reset fftw3 internal multithread capability */
    LW_FFTW_NAMES(plan_with_nthreads)(lwfftw_multhr_nthreads);
  }

#else
  lwfftw_multhr_thread_execute_func(NULL);
#endif

  for(i=0; i<num_windows; i++) {
    LW_FFTW_NAMES(destroy_plan)(plans[i]);
  }

}
