#include "lw_fftw3.h"

/* *************************
   Complex to Complex Transforms
   ************************* */
static void lwfftw_c2c_1d(SIGNAL in, SIGNAL out, int fftw_sign, unsigned fftw_flags) {
  lw_fftw_plan p;
  int logical_size = in->size/2;
  
  /* <in> represents a complexe signal -> we expect in->size to be even */
  if( 2*logical_size != in->size ) {
    Errorf("If <in> is a complex fourier transform, in->size should be even !!");
  }
 
  SizeSignal(out,in->size,YSIG);

  p = LW_FFTW_NAMES(plan_dft_1d)( logical_size,
				  (lw_fftw_complex *) in->Y,
				  (lw_fftw_complex *) out->Y,
				  fftw_sign, fftw_flags);
  
  LW_FFTW_NAMES(execute)(p);
  
  LW_FFTW_NAMES(destroy_plan)(p);
}

/* *************************
   Real to Complex Transforms
   ************************* */
static void lwfftw_r2c_1d(SIGNAL in, SIGNAL out, unsigned fftw_flags) {
  lw_fftw_plan p;

  /* out will contain the output complex results: out->size = 2*(in->size/2+1) */
  SizeSignal(out,2*(in->size/2+1),YSIG);

  p = LW_FFTW_NAMES(plan_dft_r2c_1d)(in->size, in->Y, (lw_fftw_complex *) out->Y, fftw_flags);
  
  LW_FFTW_NAMES(execute)(p);
  
  LW_FFTW_NAMES(destroy_plan)(p);
}


/* WARNING: this backward transform destroy input signal */
static void lwfftw_c2r_1d(SIGNAL in, SIGNAL out, int logical_size) {
  lw_fftw_plan p;

  /* -- in->size should verify: 2*(logical_size/2+1) = in->size and in[1] = 0 (imaginary part at null freq = 0)
        moreover if logical_size is even then we should also have in[in->size-1] = 0
        (imaginary part at Nyquist (n//2) freq = 0)
     -- out will contain the output real results: out->size = logical_size */
  SizeSignal(out,logical_size,YSIG);

  p = LW_FFTW_NAMES(plan_dft_c2r_1d)(logical_size, (lw_fftw_complex *) in->Y,  out->Y, FFTW_ESTIMATE);
  
  LW_FFTW_NAMES(execute)(p);
  
  LW_FFTW_NAMES(destroy_plan)(p);
}


static int lwfftw_c2r_1d_compute_check_logical_size (SIGNAL in, int logical_size) {
  /* -- in->size should verify: 2*(logical_size/2+1) = in->size and in[1] = 0 (imaginary part at null freq = 0)
     moreover if logical_size is even then we should also have in[in->size-1] = 0
     (imaginary part at Nyquist (n//2) freq = 0) */
  if (2*(in->size/2) != in->size) {
    Errorf("If <in> is a complex fourier transform, in->size should be even !!");
  }
  if (in->Y[1] != 0.0) {
    Errorf("If <in> is a complex fourier transform of a real signal, in->Y[1] should be 0 "
	   "(imaginary part at null freq = 0) !!");
  }
  if (logical_size <= 0) {
    if (in->Y[in->size-1] == 0.0) {
      logical_size = 2*((in->size/2)-1);
    } else {
      logical_size = 2*((in->size/2)-1)+1;
    }
  } else {
    if (2*(logical_size/2+1) != in->size) {
      Errorf("We should have 2*(logical_size//2+1) = in->size");
    }
    if ( (2*(logical_size/2) == logical_size) && (in->Y[in->size-1] != 0.0) ) {
      Errorf("If <in> is a complex fourier transform of a real signal and logical_size is even "
	     "then in->Y[in->size-1] should be 0 (imaginary part at Nyquist (logical_size/2) freq = 0) !!");
    }
  }

  return logical_size;
}

/* ************************************
   Real to Hermitian Complex Transforms
   ************************************ */
static void lwfftw_r2hc_1d(SIGNAL in, SIGNAL out) {
  lw_fftw_plan p;

  /* out is of size in->size and will contain the fft in the form:
     r_0, r_1, r_2, ..., r_{n/2}, i_{(n+1)/2-1}, ..., i_2, i_1
  */
  SizeSignal(out,in->size,YSIG);

  p = LW_FFTW_NAMES(plan_r2r_1d)(in->size, in->Y, out->Y, FFTW_R2HC, FFTW_ESTIMATE);
  
  LW_FFTW_NAMES(execute)(p);
  
  LW_FFTW_NAMES(destroy_plan)(p);
}

/* WARNING: this backward transform destroy input signal */
static void lwfftw_hc2r_1d(SIGNAL in, SIGNAL out) {
  lw_fftw_plan p;

  /* out is of size in->size and will contain the real results
   */
  SizeSignal(out,in->size,YSIG);

  p = LW_FFTW_NAMES(plan_r2r_1d)(in->size, in->Y, out->Y, FFTW_HC2R, FFTW_ESTIMATE);
  
  LW_FFTW_NAMES(execute)(p);
  
  LW_FFTW_NAMES(destroy_plan)(p);
}


/* **************************************
   Multiple Complex to Complex Transforms
   ************************************** */
static void lwfftw_multiple_c2c_1d(SIGNAL in, SIGNAL out,
				   int logical_size, int num_windows, int logical_step,
				   int fftw_sign, unsigned fftw_flags) {
  lw_fftw_plan p;

  /* input signal should contain at least (num_windows-1)*logical_step+logical_size complex numbers */
  if( in->size < 2*((num_windows-1)*logical_step+logical_size) ) {
    Errorf("lwfftw_multiple_c2c_1d(): Input signal size is too small !!");
  }

  /* out will contain the output complex results: out->size = num_windows*2*logical_size */
  SizeSignal(out, num_windows*2*logical_size, YSIG);

  p = LW_FFTW_NAMES(plan_many_dft)(1, &logical_size, num_windows,
				   (lw_fftw_complex *) in->Y, NULL, 1, logical_step,
				   (lw_fftw_complex *) out->Y, NULL, 1, logical_size,
				   fftw_sign, fftw_flags);

  if (p == NULL) {Errorf("lwfftw_multiple_c2c_1d(): Problem while making plan !!");}

  LW_FFTW_NAMES(execute)(p);
  
  LW_FFTW_NAMES(destroy_plan)(p);
}


/*
 * Command for fftw 
 */
void C_Fftw(char **argv) {
  char *action;
 
  argv = ParseArgv(argv,tWORD,&action,-1);

 
  if (!strcmp(action,"nthreads")) { /* 'nthreads' action */
    int nthreads;

    argv = ParseArgv(argv,tINT_,0,&nthreads,0);

#   if !defined(LW_FFTW_NO_THREAD) 
    if (nthreads > 0) {
      lwfftw_multhr_nthreads = nthreads;
      LW_FFTW_NAMES(plan_with_nthreads)(nthreads);
    }
    SetResultInt(lwfftw_multhr_nthreads);
#   else
    Errorf("Package FFTW3 was compiled with threads disabled.");
#   endif


  } else if (!strcmp(action,"import_wisdom")) { /* 'import_wisdom' action */
    char *fileName;
    int res;
    FILE *fp;

    argv = ParseArgv(argv,tSTR,&fileName,0);

    fp = FOpen(fileName,"r");
    if (fp == NULL) {Errorf("Can't open file: %s !!",fileName);}
    res = LW_FFTW_NAMES(import_wisdom_from_file)(fp);
    FClose(fp);
    if (res != 1) {Errorf("Error while reading/interpreting wisdom (format error ?) (%d) !!", res);}


  } else if (!strcmp(action,"export_wisdom")) { /* 'export_wisdom' action */
    char *fileName;
    FILE *fp;

    argv = ParseArgv(argv,tSTR,&fileName,0);

    fp = FOpen(fileName,"w");
    if (fp == NULL) {Errorf("Can't open file: %s !!",fileName);}
    LW_FFTW_NAMES(export_wisdom_to_file)(fp);
    FClose(fp);

  } else if (!strcmp(action,"ri2c")) { /* 'ri2c' action */
    SIGNAL in1, in2, out;
    int i,j;

    argv = ParseArgv(argv,tSIGNALI,&in1,tSIGNALI,&in2,tSIGNAL,&out,0);
    if (in1->size != in2->size) {Errorf("Input signals should have the same size !!");}

    SizeSignal(out,2*in1->size,YSIG);
    for(i=0,j=0; i<in1->size; i++,j+=2) {
      out->Y[j]= in1->Y[i];
      out->Y[j+1]= in2->Y[i];
    }

  } else if (!strcmp(action,"c2ri")) { /* 'c2ri' action */
    SIGNAL in, re, im;
    int i,j,logical_size;

    argv = ParseArgv(argv,tSIGNALI,&in,tSIGNAL,&re,tSIGNAL,&im,0);
    logical_size = in->size/2;
    if (in->size != 2*logical_size) {Errorf("Input signals should have an even size !!");}

    SizeSignal(re,logical_size,YSIG);
    SizeSignal(im,logical_size,YSIG);
    for(i=0,j=0; i<logical_size; i++,j+=2) {
      re->Y[i]= in->Y[j];
      im->Y[i]= in->Y[j+1];
    }

  } else if (!strcmp(action,"c2mp")) { /* 'c2mp' action */
    SIGNAL in, module, phase;
    int i,j,logical_size;

    argv = ParseArgv(argv,tSIGNALI,&in,tSIGNAL,&module,tSIGNAL_,NULL,&phase,0);
    logical_size = in->size/2;
    if (in->size != 2*logical_size) {Errorf("Input signals should have an even size !!");}

    SizeSignal(module,logical_size,YSIG);
    if (phase != NULL) {
      SizeSignal(phase,logical_size,YSIG);
    }

    for(i=0,j=0; i<logical_size; i++,j+=2) {
      module->Y[i] = hypot(in->Y[j],in->Y[j+1]);
      if (phase != NULL) {phase->Y[i] = atan2(in->Y[j+1],in->Y[j]);}
    }

  } else if (!strcmp(action,"cprod")) { /* 'cprod' action */
    SIGNAL in1, in2, out;
    int i, logical_size, ir, ii;
    LWFLOAT in1_re;

    argv = ParseArgv(argv,tSIGNALI,&in1,tSIGNALI,&in2,tSIGNAL,&out,0);

    if (2*(in1->size/2) != in1->size) {Errorf("If input sigs are complex fourier transform, in->size should be even !!");}
    if (in1->size != in2->size) {Errorf("Input signals should have the same size !!");}
    if (out == in2) {Errorf("<out> has to be diferent from <in2>.");}
    SizeSignal(out,in1->size,YSIG);
    logical_size = in1->size/2;
    for(i=0; i<logical_size;i++) {
      ir = 2*i;
      ii = 2*i+1;
      in1_re = in1->Y[ir];
      out->Y[ir] = in1_re*in2->Y[ir]-in1->Y[ii]*in2->Y[ii];
      out->Y[ii] = in1_re*in2->Y[ii]+in1->Y[ii]*in2->Y[ir];
    }


  } else {
    LWFLOAT time;
    SIGNAL in,out;
    char opt;
    int num_windows, logical_step, logical_size = 0;
    int fftw_sign = FFTW_FORWARD;
    unsigned fftw_flags = FFTW_ESTIMATE;
    int flag_alt = NO;

    argv = ParseArgv(argv,tSIGNALI,&in,tSIGNAL,&out,-1);
    if (!strcmp(action,"multi_c2c")) {
      argv = ParseArgv(argv,tINT,&logical_size,tINT,&num_windows,tINT_,0,&logical_step,-1);
      if( logical_step == 0 ) {logical_step = logical_size;}
    }
    
    while(opt = ParseOption(&argv)) {
	switch(opt) {
	case 'l': /* logical size */
	  if ( strcmp(action,"c2r") && strcmp(action,"multi_r2c") ) {ErrorOption(opt);}
	  argv = ParseArgv(argv,tINT,&logical_size,-1);
	  break;
	case 'a': /* alt */
	  flag_alt = YES;
	  break;
	case 'b': /* backward orientation for the transform */
	  if ( strcmp(action,"c2c") && strcmp(action,"multi_c2c") ) {ErrorOption(opt);}
	  fftw_sign = FFTW_BACKWARD;
	  break;
	case 'm': /* measure plan */
	  fftw_flags = FFTW_MEASURE;
	  break;
	case 'p': /* measure plan - patient mode */
	  fftw_flags = FFTW_PATIENT;
	  break;
	case 'x': /* measure plan - exhaustive mode */
	  fftw_flags = FFTW_EXHAUSTIVE;
	  break;
        default:
          ErrorOption(opt);
        }
      }
      NoMoreArgs(argv);


    time = MyTime();
    
    /*** 'c2c' action */
    if (!strcmp(action,"c2c")) {
      lwfftw_c2c_1d(in, out, fftw_sign, fftw_flags);
      
      /*** 'multi_c2c' action */
    } else if (!strcmp(action,"multi_c2c")) {
      if(flag_alt == YES) {
	extern void lwfftw_multhr_multiple_c2c_1d(SIGNAL in, SIGNAL out,
						  int logical_size, int num_windows, int logical_step,
						  int fftw_sign, unsigned fftw_flags);
	lwfftw_multhr_multiple_c2c_1d(in, out, logical_size, num_windows, logical_step, fftw_sign, fftw_flags);
      } else {
      	lwfftw_multiple_c2c_1d(in, out, logical_size, num_windows, logical_step, fftw_sign, fftw_flags);
      }
      /*** 'r2c' action */
    } else if (!strcmp(action,"r2c")) {
      lwfftw_r2c_1d(in,out,fftw_flags);
            
      /*** 'c2r' action */
    } else if (!strcmp(action,"c2r")) {
      logical_size = lwfftw_c2r_1d_compute_check_logical_size (in,logical_size);
      lwfftw_c2r_1d(in,out,logical_size);

      /*** 'r2hc' action */
    } else if (!strcmp(action,"r2hc")) {
      lwfftw_r2hc_1d(in,out);

      /*** 'hc2r' action */
    } else if (!strcmp(action,"hc2r")) {
      lwfftw_hc2r_1d(in,out);

    } else {
      Errorf("Unknow action '%s'",action);
    }
    
    SetResultFloat(MyTime()-time);
  }

}


static CProc fftwCommands[] = {
  "cfftw",C_Fftw,

  "{{description} {provides direct access to the low-level interface with "
  "the fftw3 library. In particular, complexe input/output signal contents "
  "follow the internal ordering of real and imaginary parts of fftw3. "
  "Moreover, all these transforms are NOT normalized (ifft(fft(sig))=sig.size*sig).\n"
  "In general, you should rather be using the 'wrapper' routines.}} "

  "{{nthreads <number of threads to use>} {If fftw3 multi-threading is enabled, "
  "sets the maximun number of threads to use by the fftw3 library.}} "

  "{{import_wisdom <wisdom file name>} {Read saved fftw wisdom from file. "
  "Shell command fftw-wisdom|fftwlwisdom can be used to generate wisdom file.}} "

  "{{export_wisdom <wisdom file name>} {Save current fftw wisdom to file. }} "

  "{{ri2c <RealPart> <ImaginaryPart> <ComplexOut>} "
  "{Combines real and imaginary parts to a complex signal where real and imaginary parts alternate: "
  " Re_i = sig[2*i] and Im_i = sig[2*i+1].}} "

  "{{c2ri <ComplexIn> <RealPart> <ImaginaryPart>} "
  "{Decompose a complex signal into real and imaginary parts.}} "

  "{{c2mp <ComplexIn> <Module> <Phase>} "
  "{Decompose a complex signal into module and phase signals.}} "

  "{{cprod <In1(complex)> <In2(complex)> <Out(complex)>} "
  "{Compute the product of two complex signals. For INPLACE product, use <out>=<in1>. Real and imaginary parts alternate: "
  " Re_i = sig[2*i] and Im_i = sig[2*i+1].}} "

  "{{c2c <signalInComplex> <signalOutComplex> [-b]} "
  "{compute the Fourier transform of a complex input signal. Real and imaginary parts alternate: "
  " Re_i = sig[2*i] and Im_i = sig[2*i+1].\n-b: compute backward transform.}} "

  "{{multi_c2c <signalInComplex> <signalOutComplex> <logical_size> <num_windows> [<logical_step> = <logical_size>] [-b]} "
  "{compute the Fourier transform of a complex input signal over windows of size <logical_size> (expressed in number of "
  "complex values) with a window step of <logical_step>. Output contains the <num_windows> fourier transforms. "
  "(Real and imaginary parts alternate: "
  "Re_i = sig[2*i] and Im_i = sig[2*i+1].\n-b: compute backward transform.)}} "

  "{{r2c <signalInReal> <signalOutComplexe>} "
  "{compute the Fourier transform of a real signals. Output size is 2*(in->size//2+1), "
  " Re_i = out[2*i] and Im_i[2*i+1].}} "

  "{{c2r <signalInComplexe> <signalOutReal> [-l <real_signal_size>]} "
  "{compute the inverse Fourier transform of a real signals. Input should be the "
  "Fourier transform of a real signal i.e. in[1] == 0 (imaginary part at null freq = 0) "
  "and in[in.size-1] == 0 when the real signal size is even (imaginary part at "
  "Nyquist (n//2) freq = 0). Real signal size is automatically determined based on "
  "in[in.size-1] == 0. In case of a real signal of odd size with in[in.size-1] == 0, you "
  "can force the output signal size using the -l option.\n"
  "WARNING: Destroys the content of the input signal.}} "

  "{{r2hc <signalInReal> <signalOutComplexe>} "
  "{compute the Fourier transform of a real signals. Output is of size in->size. "
  "Real and imaginary parts are ordered as follows:\n"
  "r_0, r_1, r_2, ..., r_{n/2}, i_{(n+1)//2-1}, ..., i_2, i_1.}} "

  "{{hc2r <signalInComplexe> <signalOutReal>} "
  "{compute the inverse Fourier transform of a real signals. The real and imaginary parts "
  "of the input signal should be ordered as follows:\n"
  "r_0, r_1, r_2, ..., r_{n/2}, i_{(n+1)//2-1}, ..., i_2, i_1.\n"
  "WARNING: Destroys the content of the input signal.}} ",

  /*  "fftwconv",C_FftwConvolByPart,
      "{{<in> <filter> <size_window> <out>} {Not documented yet.}} {{} {}} " ,
  */

  "cfftwconv",C_Lwfftw_convol,
  "{{window_size <in> <filter>} "
  "{Returns {optimal_window_size num_windows}.}} "

  "{{pad <in> <filter> <size_window> <in_padded> [-b 'pad'|'pad0'|'per'|'mir']} "
  "{Not documented yet. Returns the number of windows.}} "

  "{{filter_info <filter>} {Returns {m_neg m_pos}.}} "

  "{{perio_filter <filter_re> <filter_im>|null <window_size> <filter_perio>} "
  "{Compute the <window_size>-periodic filter. If <filter_im> is not null <filter_perio> "
  "is a complex signal. Returns filter info {m_neg m_pos}.}} "

  "{{multi_r2c <in_padded> <filter_logical_size> <window_size> <num_windows> <ft_in_padded>} "
  "{Compute the FTs of input signals. Returns <window_size_fft>.}} "

  "{{complex_prod <ft_in_padded> <ft_filter> <num_windows> [<prod_ft>]} "
  "{Compute the product of the multiple FT <ft_in_padded> with the FT of the filter. "
  "If you don't provide <prod_ft> product is done IN PLACE <ft_in_padded>.}} "

  "{{multi_c2r <prod_ft> <filter> <window_size> <num_windows> <index_first_good> <size_good> "
  "<conv_sig> [<norm>=1.0]} {Compute the multiple inverse <window_size>-FTs. <filter>, "
  "<index_first_good> and <size_good> are used to reconstruct the different part of the result "
  "signal <conv_sig>.}} "

  "{{recons_c2mp <complex_in> <filter> <window_size> <num_windows> <index_first_good> <size_good> "
  "<module_sig> [phase_sig] [<norm>=1.0]} {<complex_in> is a multi-part complex convolution signals. <filter>, "
  "<index_first_good> and <size_good> are used to identify the different part of the result. "
  "Returns the module <module_sig> and phase <module_sig> (if provided) of the convolution.}} ",

  NULL,NULL,NULL
};

static CProcTable fftwTable = 
{fftwCommands,"fftw3",
 "Commands related to the fftw3 library"};


/***********************************
 * 
 * Loading/Adding the fftw3 package
 *
 ***********************************/

static void LoadFftw3Package(void) {

# if !defined(LW_FFTW_NO_THREAD) 
  if (LW_FFTW_NAMES(init_threads) == 0) {
    Errorf("FFTW3 package: Error initialising threads !!! (Weird)");
  }
# endif

  AddCProcTable(&fftwTable);
}

void DeclareFftw3Package(void) {
  DeclarePackage("fftw3",LoadFftw3Package,2007,"1.0","B.Audit",
		 "Package allowing to use the fftw3 library. "
		 "Fftw library was developed at MIT by Matteo "
		 "Frigo and Steven G. Johnson. "
		 "(http://www.fftw.org/)"
		 );
}

