/*..........................................................................*/
/*                                                                          */
/*      Pollux 2.1                                                          */
/*                                                                          */
/*      Authors: Benjamin Audit, Emmanuel Bacry and Cedric Vaillant         */
/*      Copyright (C) 1997-2005.  All Right Reserved.                       */
/*                                                                          */
/*      Contact: Benjamin.Audit@ens-lyon.fr                                 */
/*                                                                          */
/*..........................................................................*/
/*                                                                          */
/*      This program is a free software, you can redistribute it and/or     */
/*      modify it under the terms of the GNU General Public License as      */
/*      published by the Free Software Foundation; either version 2 of the  */
/*      License, or (at your option) any later version                      */
/*                                                                          */
/*      This program is distributed in the hope that it will be useful,     */
/*      but WITHOUT ANY WARRANTY; without even the implied warranty of      */
/*      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the       */
/*      GNU General Public License for more details.                        */
/*                                                                          */
/*      You should have received a copy of the GNU General Public License   */
/*      along with this program (in a file named COPYRIGHT);                */
/*      if not, write to the Free Software Foundation, Inc.,                */
/*      59 Temple Place, Suite 330, Boston, MA  02111-1307  USA             */
/*                                                                          */
/*..........................................................................*/

#include "lastwave.h"
#include "signals.h"
#include "wtrans1d.h"

/* The Logarithm of the Gamma function */
static double LnGamma(double xx,int *pSign)
{
  double x,tmp,ser;
  static double cof[6]={76.18009173,-86.50532033,24.01409822,
			-1.231739516,0.120858003e-2,-0.536382e-5};
  int j;
  
  /* Different Cases */
  *pSign = 1;
  if (xx <= 0) {
    if (sin(M_PI*xx) < 0) *pSign = -1;
    return(log(M_PI) - LnGamma(1-xx,&j) - log(fabs(sin(M_PI*xx))));
  }	
  
  x=xx-1.0;
  tmp=x+5.5;
  tmp -= (x+0.5)*log(tmp);
  ser=1.0;
  for (j=0;j<=5;j++) {
    x += 1.0;
    ser += cof[j]/x;
  }
  return -tmp+log(2.50662827465*ser);
}

/* Computes the C(n,k) coefficients (k integer, n float) */
static double Cnk(double n, int k)
{
  int sign1,sign2,sign3;
  double res1;
  int j;
  
  if (k == 0) return (1.);
  if (n >= 0. && (long) n == n && k > n)
    return (0.);
    
  res1 = exp( LnGamma(n+1,&sign1) - LnGamma(n-k+1,&sign2) - LnGamma(k+1,&sign3));
  res1 = sign1*sign2*sign3*res1;
  
  return(res1);
}

#define PRECCFI 1.e-6
static int CausalFracIntSideEffect(double s)
{
  static double sRef = 1.0;
  static int resRef = 1;

  double sum = 0.;
  double cnk2;
  int i=0;

  if (s == sRef)
    return(resRef);
  
  sRef=s;
  s = 1-s;
  while(1)
    {
      cnk2 = Cnk(s,i);
      cnk2 = cnk2*cnk2;
      sum += cnk2;
      if (cnk2 < PRECCFI*sum)
	return(resRef=i);
      i++;
    }
}

/* Computes the C(n,k) coefficients and put them in a signal 
 * k is an integer and n is a (positive or negative) float */
static void CnkSignal(SIGNAL s,double n,int longueur)
{
  int i;

  SizeSignal(s,longueur,YSIG);
  s->x0 = 0.;
  for(i=0;i<longueur;i++)
    s->Y[i] = Cnk(n,i);
  
}

/* Conjugate a signal : (-1)^k*x(k) */
static void ConjSig(SIGNAL signal)
{
  int sign,i;

  if (((int) (fabs(signal->x0) + 0.5))%2 == 0) sign = 1;
  else sign = -1;

  for (i=0;i<signal->size;i++) 
    {
      signal->Y[i] *= sign;
      sign = -sign;
    }
}

static void CausalFracInt(SIGNAL sigin,SIGNAL sigout,double s)
{
  static double sRef = 0;
  static int sizeRef = 0;
  static SIGNAL signal = NULL;
  static SIGNAL signalTemp = NULL;

  int i,j;
  
  /* Some inits if necessary */
  if (signal == NULL) 
    signal = NewSignal();

  if (signalTemp == NULL) 
    signalTemp = NewSignal();

  if(sigin->size != sizeRef || s != sRef)
    {
      sizeRef = sigin->size;
      sRef = s;
      CnkSignal(signalTemp,-s+1.,sizeRef);
      ConjSig(signalTemp);
      SizeSignal(signal,sizeRef,YSIG);
      
      signal->Y[0] = 0.;
      for(i=1;i<sizeRef-1;i++)
	signal->Y[i] = signalTemp->Y[i-1];
      signal->Y[sizeRef-1] = 0.;
    }
  
  SizeSignal(sigout,sizeRef,YSIG);

  signalTemp->Y[0] = sigin->Y[0];
  for(i=1;i<sizeRef;i++)
    signalTemp->Y[i] = signalTemp->Y[i-1]+sigin->Y[i];  

  cv_flt_init_n(CV_RC_FORM,sizeRef,0,0,0,signalTemp->Y,NULL);
  cv_sig_init(CV_RC_FORM,signal->Y,NULL,sizeRef);
  cv_set_method(CV_UNDEFINED);
  cv_compute(CV_PADDING,sigout->Y,&i,&j);
}

void Farima(SIGNAL signal,int size,FLOAT H,FLOAT sigma) {
  static int sizeRef = 0;
  static FLOAT HRef = 0.;
  static int offset = 0;
  static SIGNAL sig_noise = NULL;
  static SIGNAL sig_farima = NULL;

  int i;

  if( size <= 0 || H<=0. || H>=1. || sigma <= 0.)
    Errorf("Farima(): One of the parameter is not in bound. "
	   "(size>0, 0<H<1, sigma>0)");

  /* Some init if necessary */
  if( sig_noise == NULL ) {sig_noise = NewSignal();}
  if( sig_farima == NULL ) {sig_farima = NewSignal();}
  if( size != sizeRef || H != HRef ) {
    sizeRef = size;
    HRef = H;
    offset = CausalFracIntSideEffect((double) (H+0.5));
    SizeSignal(sig_noise,sizeRef+offset,YSIG);
  }

  /* do the computation */
  sigma = sigma*exp(LnGamma(2.*H+1.,&i))*sin(M_PI*H);

  for(i=0; i < sig_noise->size; i++)
    sig_noise->Y[i] = Grand(sigma);
	
  CausalFracInt(sig_noise,sig_farima,(double) (H+0.5));
	
      
  SizeSignal(signal,size,YSIG);

  for(i=0;i<signal->size;i++)
    signal->Y[i] = sig_farima->Y[i+offset];
}

void Farima2H(WTRANS wtrans,SIGNAL signal,int noct,int size,
	      FLOAT h,FLOAT H,
	      FLOAT sigma,char *filename)
{
  int o;
  int i,offset;
  int p2j,sizeJ,sizeFilterH,maxSupport;
  FLOAT factor,s;
  
  if( size <= 0 || noct <= 0 
      || h<=0. || h>=1. || H<=0. || H>=1.
      || sigma <= 0.)
    Errorf("Farima(): One of the parameter is not in bound. "
	   "(size>0, 0<h<1, 0<H<1, sigma>0)");

  /* if filename != NULL we do the init */
  if (filename != NULL)
    {
      if (!INRANGE(0,noct,NOCT-1)) 
	Errorf("Bad range for <noct>");
      
      SetBiorFG(filename,wtrans);
      maxSupport = MAX(RIGHT_FILT(wtrans->fg->H2),RIGHT_FILT(wtrans->fg->G2))
	-MIN(LEFT_FILT(wtrans->fg->H2),LEFT_FILT(wtrans->fg->G2)) + 1;
      sizeFilterH = wtrans->fg->H2->size;
      
      p2j = 1 << (noct-1);
      sizeJ = size + p2j*maxSupport + (p2j-1)*sizeFilterH;
      p2j *=2;
      sizeJ = sizeJ/p2j + ((sizeJ%p2j == 0) ? 0 : 1);
      /*
	Printf("%d %d %d\n",wtrans->fg->H2->size,LEFT_FILT(wtrans->fg->H2),
	RIGHT_FILT(wtrans->fg->H2));
	Printf("%d %d %d\n",wtrans->fg->G2->size,LEFT_FILT(wtrans->fg->G2),
	RIGHT_FILT(wtrans->fg->G2));
	Printf("%d\n",sizeJ);
      */
      
      wtrans->nOct = noct;
      wtrans->nVoice = 1;
      wtrans->type = W_ORTH;
      
      SizeSignal(wtrans->A[noct][0],sizeJ,YSIG);

      for(o = noct;o >= 1;o--)
	{
	  SizeSignal(wtrans->D[o][0],sizeJ,YSIG);
	  sizeJ *= 2;
	}

    }

  /* else we do the computation */
  else
    {
      s = h + 0.5;
      
      /* The approximation at scale noct is a Farima H 
	 of stddev sigma*2^(noct(sqrt(H*h)+0.5)) 
	 (why not h+0.5 or H+0.5 or (H+h)/2 ????)
       */
      Farima(wtrans->A[noct][0],wtrans->A[noct][0]->size,H,
	     sigma*pow(2,noct*(sqrt(h*H)+0.5)));
      
      /* Details are white noise of appropriate:-) stddev */
      sigma = sigma*pow(2,noct*s)*exp(LnGamma(2.*h+1.,&i))*sin(M_PI*h);
      sigma = sigma*pow(4.,-s);
      factor = pow(2.,-s);
      for(o = noct;o >= 1;o--)
	{
	  for(i=0; i < wtrans->D[o][0]->size; i++)
	    wtrans->D[o][0]->Y[i] = Grand(sigma);
	  sigma *= factor;
	}
      OWtr(wtrans,wtrans->A[0][0]);

      offset = wtrans->A[0][0]->lastp-size+1;
      if(offset<0) {Errorf("Farima2H(): Weird !!");}
      
      SizeSignal(signal,size,YSIG);
      for(i=0;i<size;i++)
	signal->Y[i] = wtrans->A[0][0]->Y[i+offset];
    }
}
