/*..........................................................................*/
/*                                                                          */
/*      Pollux 2.1                                                          */
/*                                                                          */
/*      Authors: Benjamin Audit, Emmanuel Bacry and Cedric Vaillant         */
/*      Copyright (C) 1997-2005.  All Right Reserved.                       */
/*                                                                          */
/*      Contact: Benjamin.Audit@ens-lyon.fr                                 */
/*                                                                          */
/*..........................................................................*/
/*                                                                          */
/*      This program is a free software, you can redistribute it and/or     */
/*      modify it under the terms of the GNU General Public License as      */
/*      published by the Free Software Foundation; either version 2 of the  */
/*      License, or (at your option) any later version                      */
/*                                                                          */
/*      This program is distributed in the hope that it will be useful,     */
/*      but WITHOUT ANY WARRANTY; without even the implied warranty of      */
/*      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the       */
/*      GNU General Public License for more details.                        */
/*                                                                          */
/*      You should have received a copy of the GNU General Public License   */
/*      along with this program (in a file named COPYRIGHT);                */
/*      if not, write to the Free Software Foundation, Inc.,                */
/*      59 Temple Place, Suite 330, Boston, MA  02111-1307  USA             */
/*                                                                          */
/*..........................................................................*/

/****************************************************************************/
/*                                                                          */
/*      FUNCTIONS TO DEAL WITH POLLUX STRUCTURE                            */
/*                                                                          */
/****************************************************************************/



#include "pollux.h"
#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#ifndef NOT_USED_WITH_LW

extern void *Malloc(size_t size);
extern void *Calloc(int n, size_t size);
extern void Free(void * ptr);

#define malloc Malloc
#define calloc Calloc
#define free Free

#endif /* NOT_USED_WITH_LW */

/****************************************************************************/
/* SOME USEFUL LITTLE FUNCTIONS                                             */
/****************************************************************************/

/****************************************************************************/
/* RETURNS THE POINTER TO THE FIRST NON BLANK CARACTERE OF A STRING          */
/****************************************************************************/
char *RemoveLeadBlank( char *string)
{
  while( isspace(*string) ) string++;

  return string;
}
/****************************************************************************/
/* REPLACES ALL TRAILING BLANKS IN A STRING WITH '\0'                       */
/****************************************************************************/
void RemoveTrailBlank( char *string)
{
  int size;

  for(size = strlen(string) - 1;(size>=0) && isspace(string[size]);size--)
      string[size] = '\0';
}

/****************************************************************************/
/*   INIT ALL LINES IN A POLLUX LIST TO \0                                  */
/*                                                                          */
/****************************************************************************/
/*                                                                          */
/*                                                                          */
/****************************************************************************/
void PolInitList( char polList[][POLSTRSIZE])
{
  int i;

  for(i=0;i<POLSIZELIST;i++)
    {
      polList[i][0] = '\0';
    }
}



/****************************************************************************/
/* THIS FUNCTION RETURNS POLYES IF string IS PART OF THE stringList         */
/* IT RETURNS POLNO OTHERWISE                                               */
/* IT ALSO SET pPos TO THE INDEX WHERE STING WAS FOUND                      */
/****************************************************************************/
/* IT ASSUMES THAT stringList[i] IS A STRING (IT MUST BE TERMINATED BY \0)  */
/* AND THAT THE stringList IS TERMINATED BY AN EMPTY STRING                 */
/* (stringList[i][0] = '\0')                                                */
/****************************************************************************/
int IsOfStringList(const char *string, const char stringList[][POLSTRSIZE], int *pPos)
{
  int i;

  for(i=0;(i < POLSIZELIST) && (stringList[i][0] != '\0');i++)
    {
      if(!strcmp(stringList[i],string)) 
	{
	  *pPos = i;
	  return POLYES;
	}
    }
  return POLNO;
}
/****************************************************************************/
/* COPY A STRING INTO AN OTHER ONE VERIFYING THE ORIGINAL STRING FITS IN A  */
/* PolString                                                                */
/****************************************************************************/
/****************************************************************************/

int PolStrCpy(char *string,const char *value)
{
  if(strlen(value) >= POLSTRSIZE)
    return POLOVERFLOW;
  
  strcpy(string,value);
  
  return POLOK;
}
/****************************************************************************/
/* CONCAT A STRING TO AN OTHER ONE VERIFYING THERE IS ENOUGH PLACE TO FIT   */
/* IN A PolString                                                                */
/****************************************************************************/
/****************************************************************************/

int PolStrCat(char *string,const char *value)
{
  if(strlen(string)+strlen(value) >= POLSTRSIZE)
    return POLOVERFLOW;
  
  strcat(string,value);
  
  return POLOK;
}

/****************************************************************************/
/* LOWER A STRING                                                           */
/****************************************************************************/
/****************************************************************************/

char *PolStrLower(char *string)
{
  char *c= string;

  while( *c != '\0')
    {
      *c = tolower( (unsigned char) *c);
      c++;
    }
  
  return string;
}

/****************************************************************************/
/* CONCATENATES A LINE AT THE END OF A BUFFER                               */
/* IT VERIFY THAT ENOUGH SPACE IS AVAILABLE                                 */
/****************************************************************************/
int PolLineToBuffer(PolluxBuffer *polbuffer, char *line)
{
  int sizeb,sizel;
  sizeb=polbuffer->size;
  sizel=strlen(line);
  if(sizeb + sizel < POLBUFFERSIZE)
    {
      strcat(polbuffer->buffer+sizeb,line);
      polbuffer->size = sizeb+sizel;
      return POLOK;
    }
  else return POLOVERFLOW;
}

/****************************************************************************/
/* CONCATENATES field AND value TO MAKE A DEFINITION LINE IN line           */
/* IT VERIFY THAT ENOUGH SPACE IS AVAILABLE                                 */
/****************************************************************************/
/****************************************************************************/
int PolFieldValueToLine(char *line, char *field, char *value)
{
  RemoveTrailBlank(value);

  if(strlen(field) + strlen(value) + 3 < POLSTRSIZE)
    {
      sprintf(line,"=%s %s\n",field,value);
      return POLOK;
    }
  else return POLOVERFLOW;
}

/****************************************************************************/
/* RETURNS POLYES IF TYPE IS OF CODING TYPE POLNO IN ALL OTHER CASES        */
/*                                                                          */
/****************************************************************************/
/* IT ASSUMES THAT LEADIND AND TRAILING BLANKS HAVE BEEN REMOVED            */
/****************************************************************************/
int IsOfCodingType(const char *type)
{
  if(!strcmp(type,"cds") || !strcmp(type,"operon"))
    return POLYES;
  else return POLNO;
}

/****************************************************************************/
/* IT TRANSFORMS A DNA SIGNAL (CONSISTING OF +1, -1 AND 0) INTO AN ARRAY OF */
/* INTEGERS, IT RETURNS THE SIZE OF THE ARRAY OR A NEGATIVE VALUE IF IT     */
/* FAILED                                                                   */
/****************************************************************************/
/* malloc IS USED TO GET THE SPACE FOR THE ARRAY, IT MUST BE FREE WITH free */
/* IT ASSUMES THAT signal->size is greater (or equal) than 1                */
/****************************************************************************/
int *AdnSigToInt(FLOAT *signal, int size)
{
  int *intSignal;
  int i;
  FLOAT eps = 0.0001;

  intSignal = (int *) ( malloc(sizeof(int)*size));
  if( intSignal == NULL )
    return NULL;
    
  for(i=0;i < size;i++)
    {
      if(abs(signal[i] - 1) < eps)
	intSignal[i] = 1;
      else if(abs(signal[i] + 1) < eps)      
	intSignal[i] = -1;
      else if(abs(signal[i]) < eps)      
	intSignal[i] = 0;
      else 
	{
	  free(intSignal);
	  return NULL;
	}
    }

  return intSignal;
}
/****************************************************************************/
/* IT COMPUTES THE DERIVATIVE OF AN INTEGER SIGNAL                          */
/****************************************************************************/
/* THE SIGNAL MUST HAVE A SIZE A LEAST 2                                    */
/* size MUST BE THE SIZE OF THE SIGNAL                                      */
/****************************************************************************/
int DerIntSig(int *intSignal, int size)
{
   int i;

   if( size < 2 )
     return POLERROR;
   for(i=0;i<size-1;i++)
     {
       intSignal[i] = intSignal[i+1] - intSignal[i];
     }
   return POLOK;
}
/****************************************************************************/
/* IT RETURNS THE SPECIES PATH AND THE NUM OF A FILE PATH                   */
/* IT IS VERY USEFUL IN MACROS                                              */
/****************************************************************************/
/* "speciesPath/num": THIS RECONSTRUCTION ALWAYS GIVE THE RIGHT PATH  */
/****************************************************************************/

int PolFileToSpeciesNum( const char *filePath, char *speciesPath, char *num)
{
  int size,i;

  size = strlen(filePath);
  for(i=size-1;(i>=0) && (filePath[i] != '/');i--);
  /* NOW i IS THE INDEX OF THE LAST '/' */  

  /* num IS EVERYTHING AFTER THE LAST '/' */
  strcpy(num,(filePath + i + 1));

  strcpy(speciesPath,filePath);
  /* IF filePath IS OF THE FORM blabla/blu i>0 SO speciesPath IS "blabla" */
  /* IF filePath IS OF THE FORM /blu i=0 SO speciesPath IS "" */
  if(i>=0)
    speciesPath[i] = '\0';
  else if( i == -1 )
    {/* IF filePath IS OF THE FORM blu i=-1 SO speciesPath IS "." */
      speciesPath[0] = '.';
      speciesPath[1] = '\0';
    }
  else return POLERROR;

  return POLOK;
}
