/*..........................................................................*/
/*                                                                          */
/*      Pollux 2.1                                                          */
/*                                                                          */
/*      Authors: Benjamin Audit, Emmanuel Bacry and Cedric Vaillant         */
/*      Copyright (C) 1997-2002.  All Right Reserved.                       */
/*                                                                          */
/*      Contact: Benjamin.Audit@ena-lyon.fr                                 */
/*                                                                          */
/*..........................................................................*/
/*                                                                          */
/*      This program is a free software, you can redistribute it and/or     */
/*      modify it under the terms of the GNU General Public License as      */
/*      published by the Free Software Foundation; either version 2 of the  */
/*      License, or (at your option) any later version                      */
/*                                                                          */
/*      This program is distributed in the hope that it will be useful,     */
/*      but WITHOUT ANY WARRANTY; without even the implied warranty of      */
/*      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the       */
/*      GNU General Public License for more details.                        */
/*                                                                          */
/*      You should have received a copy of the GNU General Public License   */
/*      along with this program (in a file named COPYRIGHT);                */
/*      if not, write to the Free Software Foundation, Inc.,                */
/*      59 Temple Place, Suite 330, Boston, MA  02111-1307  USA             */
/*                                                                          */
/*..........................................................................*/

#include "pollux.h"
#include "helix_3d.h"
#include "lastwave.h"
#include "signals.h"
#include "wtrans1d.h"

#include <stdlib.h>


/* in farima.c */
extern void Farima(SIGNAL signal,int size,FLOAT H,FLOAT sigma);
extern void Farima2H(WTRANS wtrans,SIGNAL signal,int noct,int size,
		     FLOAT h,FLOAT H,FLOAT sigma,char *filename);

typedef struct courb {
  char nuc;
  FLOAT courb;
} courb;

static int comp_courb(const void *pcourb1,const void *pcourb2)
{
  courb *p1 = (courb *) pcourb1;
  courb *p2 = (courb *) pcourb2;

  if (p1->courb < p2->courb)
    return -1;
  if (p1->courb == p2->courb)
    return 0;
  if (p1->courb > p2->courb)
    return 1;
}

static void sort_tab_courb(courb tab_courb[4])
{
  qsort(tab_courb,4,sizeof(courb),comp_courb);
}

static void next_possible_courb(FLOAT tab[4][4][4], char dinu[2],
				courb tab_courb[4])
{
  int i,j,k,l;

  i = nuc_2_index(dinu[0]);
  j = nuc_2_index(dinu[1]);

  for(l=0;l<4;l++)
    {
      k = nuc_2_index(tab_courb[l].nuc);
      tab_courb[l].courb = tab[i][j][k];
    }
}

static void trinu_possible_courb(const FLOAT tab[4][4][4],
				 const char trinu[3],courb tab_courb[4])
{
  int i,j,k,l;

  i = nuc_2_index(trinu[0]);
  k = nuc_2_index(trinu[2]);

  for(l=0;l<4;l++)
    {
      j = nuc_2_index(tab_courb[l].nuc);
      tab_courb[l].courb = tab[i][j][k];
    }
}

static courb *next_courb(courb tab_courb[4],FLOAT incr_val,FLOAT courb_ref_val)
{
  int i,j;
  
  sort_tab_courb(tab_courb);
  
  /* Printf("%.6f %.6f %.6f %.6f\n",tab_courb[0].courb,tab_courb[1].courb,
     tab_courb[2].courb,tab_courb[3].courb); */

  if(incr_val > 0.)
    {
      /* WE WANT A POSITIVE INCREMENT */
      /* 1) WE LOOK FOR THE FIST INDEX GIVING A POSITIVE INCREMENT */
      for(i=0;i<4;i++)
	{
	  if(tab_courb[i].courb > courb_ref_val)
	    break;
	}
      /* 2) IF NONE EXIST WE TAKE THE BIGGER ONE */
      if(i == 4)
	return tab_courb+3;
      /* 3) ELSE CHOOSE ONE AT RANDOM */
      j = (int) (Urand()*(4-i)) + i;
      return tab_courb+j;
    }

  if(incr_val < 0.)
    {
      /* WE WANT A NEGATIVE INCREMENT */
      /* 1) WE LOOK FOR THE LAST INDEX GIVING A NEGATIVE INCREMENT */
      for(i=3;i>=0;i--)
	{
	  if(tab_courb[i].courb < courb_ref_val)
	    break;
	}
      /* 2) IF NONE EXIST WE TAKE THE SMALLER ONE */
      if(i == -1)
	return tab_courb;
      /* 3) ELSE CHOOSE ONE AT RANDOM 
	 (we choose a random integer between 0 and i included) */
      j = (int) (Urand()*(i+1));
      return tab_courb+j;
    }

  if(incr_val == 0.)
    {
      /* WE WANT A NULL INCREMENT */
      /* WE LOOK FOR THE CLOSEST ONE */
      for(i=0;i<4;i++)
	{
	  if(tab_courb[i].courb > courb_ref_val)
	    break;
	}
      if(i == 4)
	return tab_courb+3;
      if(i == 0)
	return tab_courb;
      if( (tab_courb[i].courb-courb_ref_val) < (courb_ref_val-tab_courb[i-1].courb) )
	return tab_courb+i;
      return tab_courb+i-1;
    }
}

int courb_one_h(PolluxBuffer *seqBuffer,FLOAT H,int size,FLOAT courb_ref_val)
{
  FLOAT tab[4][4][4];
  courb tab_courb[4];
  SIGNAL signal = TNewSignal();
  int i;

  /* SOME CONSISTENCY CHECKING */
  if(H>=1.0 || H <= 0. || size >= POLBUFFERSIZE || size < 3)
    return -1;

  /* INITIALISATION  */
  if (compute_courb_table(tab) != 0)
    return -1;
  tab_courb[0].nuc = 'g';
  tab_courb[1].nuc = 't';
  tab_courb[2].nuc = 'a';
  tab_courb[3].nuc = 'c';

  seqBuffer->buffer[0] = '\0'; 
  seqBuffer->size = 0;
  
  /* THE TWO FIRST NUCLEOTIDES ARE CHOSEN AT RANDOM */
  seqBuffer->buffer[0] = index_2_nuc((int) (4*Urand()));
  seqBuffer->buffer[1] = index_2_nuc((int) (4*Urand()));
  seqBuffer->buffer[2] = '\0';
  seqBuffer->size = 2;

  /* WE BUILD A FARIMA OF SIZE size-1 */
  Farima(signal,size-1,H,1.0);

  /* NOW WE LOOP FOR THE OTHER NUC */
  for(i=2;i<size;i++)
    {
      next_possible_courb(tab,seqBuffer->buffer+i-2,tab_courb);      
      seqBuffer->buffer[i] = (next_courb(tab_courb,signal->Y[i-1]-signal->Y[i-2],
					courb_ref_val))->nuc;
    }
  seqBuffer->buffer[size] = '\0';
  seqBuffer->size = size;
  
  return 0;
}

static FLOAT inverse_erf(double c)
{
  /* The relative precision is going to be of the order
     of FLT_EPSILON
  */
  int num_iter = ceil(-log(FLT_EPSILON)/log(2.));
  double min,max;
  double res,y;
  int i;
  
  if( c <= 0. )
    return 0.;
  
  if( c >= 1. )
    return FLOAT_MAX;

  /* Look for the interval to work in */
  max = 1.;
  while(erf(max)< c && max<=DBL_MAX/2.) {max*=2;}
  min = max/2.;

  if(min>=FLOAT_MAX)
    return FLOAT_MAX;

  for(i=0;i<num_iter;i++) {
    res = 0.5*(min+max);
    y = erf(res);
    if(y>c)
      max = res;
    else
      min =res;
  } 
  res = 0.5*(min+max);

  if(res>=FLOAT_MAX)
    return FLOAT_MAX;
  
  return (FLOAT) res;
}

int mononuc_one_h(PolluxBuffer *seqBuffer,FLOAT H,int size,
		  char *set1,char *set2,FLOAT c1)
{
  SIGNAL signal = TNewSignal();
  int i,size_set1,size_set2;
  FLOAT threshold,p;

#ifdef BENDEBUG
  int i1=0,i2=0;
#endif

  /* INITIALISATION  */
  seqBuffer->buffer[0] = '\0'; 
  seqBuffer->size = 0;
  size_set1 = strlen(set1);
  size_set2 = strlen(set2);

  /* SOME CONSISTENCY CHECKING */
  if( H>=1.0 || H <= 0. 
      || size >= POLBUFFERSIZE || size < 1
      || c1 <= 0. || c1 >= 1
      || size_set1 == 0 || size_set2 == 0)
    return -1;
  
  /* the threshold */
  if (c1>=0.5) {
    threshold = M_SQRT2*inverse_erf( (double)(2.*(c1-0.5)) );
  } else {
    threshold = -M_SQRT2*inverse_erf( (double)(2.*(0.5-c1)) );
  }

#ifdef BENDEBUG
  Printf("%f\n",threshold);
#endif

  /* WE BUILD A FARIMA OF SIZE size+1 */
  Farima(signal,size+1,H,1.0);

  /* WE LOOP ON NUC */
  for(i=0;i<size;i++)
    {
      do {p=Urand();} while(p>=1.0);
      if( (signal->Y[i+1]-signal->Y[i]) < threshold) {
#ifdef BENDEBUG
	i1++;
#endif
	seqBuffer->buffer[i] = set1[(int) (size_set1*p)];
      } else {
#ifdef BENDEBUG
	i2++;
#endif
	seqBuffer->buffer[i] = set2[(int) (size_set2*p)];
      }
    }
  seqBuffer->buffer[size] = '\0';
  seqBuffer->size = size;

#ifdef BENDEBUG
  Printf("i1=%d, i2=%d\n",i1,i2);
#endif

  return 0;
}

int mononuc_two_h(PolluxBuffer *seqBuffer,
		      FLOAT h, FLOAT H,int noct,int size,
		      char *set1,char *set2,FLOAT c1,char *filename)
{
  WTRANS wtrans = NewWtrans();
  TempPtr(wtrans);
  SIGNAL signal = TNewSignal();
  int i,size_set1,size_set2;
  FLOAT threshold,p;

#ifdef BENDEBUG
  int i1=0,i2=0;
#endif

  /* INITIALISATION  */
  seqBuffer->buffer[0] = '\0'; 
  seqBuffer->size = 0;
  size_set1 = strlen(set1);
  size_set2 = strlen(set2);

  /* SOME CONSISTENCY CHECKING */
  if( h>=1.0 || h <= 0. || H>=1.0 || H <= 0. 
      || size >= POLBUFFERSIZE || size < 1
      || c1 <= 0. || c1 >= 1
      || size_set1 == 0 || size_set2 == 0)
    return -1;
  
  /* the threshold */
  if (c1>=0.5) {
    threshold = M_SQRT2*inverse_erf( (double)(2.*(c1-0.5)) );
  } else {
    threshold = -M_SQRT2*inverse_erf( (double)(2.*(0.5-c1)) );
  }

#ifdef BENDEBUG
  Printf("%f\n",threshold);
#endif

  /* WE BUILD A FARIMA2H OF SIZE size+1 */
  Farima2H(wtrans,signal,noct,size+1,h,H,1.0,filename);
  Farima2H(wtrans,signal,noct,size+1,h,H,1.0,NULL);

  /* WE LOOP ON NUC */
  for(i=0;i<size;i++)
    {
      do {p=Urand();} while(p>=1.0);
      if( (signal->Y[i+1]-signal->Y[i]) < threshold) {
#ifdef BENDEBUG
	i1++;
#endif
	seqBuffer->buffer[i] = set1[(int) (size_set1*p)];
      } else {
#ifdef BENDEBUG
	i2++;
#endif
	seqBuffer->buffer[i] = set2[(int) (size_set2*p)];
      }
    }
  seqBuffer->buffer[size] = '\0';
  seqBuffer->size = size;

#ifdef BENDEBUG
  Printf("i1=%d, i2=%d\n",i1,i2);
#endif

  return 0;
}

int mononuc_two_h_old(PolluxBuffer *seqBuffer,FLOAT h,FLOAT H,int size_w,int num_repeat,
		  char *set1,char *set2,FLOAT c1, FLOAT dc1)
{
  SIGNAL signal = TNewSignal();
  SIGNAL signal_ref_val = TNewSignal();
  int i,n,size_set1,size_set2;
  FLOAT threshold,c,p;
  
  /* INITIALISATION  */
  seqBuffer->buffer[0] = '\0'; 
  seqBuffer->size = 0;
  size_set1 = strlen(set1);
  size_set2 = strlen(set2);
  
  /* SOME CONSISTENCY CHECKING */
  if( h>=1.0 || h<=0. 
      || H>=1.0 || H <= 0. 
      || size_w*num_repeat >= POLBUFFERSIZE || size_w < 1
      || c1 <= 0. || dc1 <= 0. || c1-dc1<=0. || c1+dc1>=1.
      || size_set1 == 0 || size_set2 == 0
      )
    return -1;
  
  /* WE BUILD A FARIMA(H) OF SIZE num_repeat+1 */
  Farima(signal_ref_val,num_repeat+1,H,1.0);
  
  /* WE GET THE INCREMENT */
  for(n=0;n<num_repeat;n++)
    signal_ref_val->Y[n]=signal_ref_val->Y[n+1]-signal_ref_val->Y[n];
  signal_ref_val->size = num_repeat;
  
  
  /* NOW WE LOOP ON num_repeat */
  i=0;
  for(n=0;n<num_repeat;n++) {
    /* the threshold of this widow */
    /* c = (signal_ref_val->Y[n]>0.)?c1+dc1:c1-dc1; */
    c = c1 + dc1*signal_ref_val->Y[n];
    if (c>=0.5) {
      threshold = M_SQRT2*inverse_erf( (double)(2.*(c-0.5)) );
    } else {
      threshold = -M_SQRT2*inverse_erf( (double)(2.*(0.5-c)) );
    }

    /* WE BUILD A FARIMA(h) OF SIZE size+1 */
    Farima(signal,size_w+1,h,1.0);
      
    /* NOW WE LOOP ON THE NUC */
    for(;i<(n+1)*size_w;i++) {
      do {p=Urand();} while(p>=1.0);
      if( (signal->Y[i+1-n*size_w]-signal->Y[i-n*size_w]) < threshold) {
	seqBuffer->buffer[i] = set1[(int) (size_set1*p)];
      } else {
	seqBuffer->buffer[i] = set2[(int) (size_set2*p)];
      }
    }
  }
  seqBuffer->buffer[num_repeat*size_w] = '\0';
  seqBuffer->size = num_repeat*size_w;
  
  return 0;
}
