#include "lastwave.h"
#include "signals.h"
#include "images.h"



/**************************************/
/* Compute the 'nr*nc' branches histogram */
/* of the signals Y's.                 */
/**************************************/

/*
 * input_r    : the signal to make the histogram along the rows 
 * input_c    : the signal to make the histogram along the columns 
 * output    : the image to put the histogram into
 * n_r, n_c    : the numbers of branches of the histogram
 * ymin_r, ymax_r, ymin_c, ymax_rc :
               the yMin and yMax of the histogram (not used if min > max)
 * xmin_r, xmax_r, xmin_c, xmax_c :
               the xMin and xMax between which the histogram is made
 * weight   : signal of weights (or NULL). It must have the same size as input 
 */
 
void Histo2Sig(IMAGE output,
	       SIGNAL input_r, int n_r, FLOAT xmin_r, FLOAT xmax_r, FLOAT ymin_r, FLOAT ymax_r,
	       SIGNAL input_c, int n_c, FLOAT xmin_c, FLOAT xmax_c, FLOAT ymin_c, FLOAT ymax_c,
	       int flagCausal)
{
  int i,index_r,index_c;
  FLOAT min,max;
  int iMin,iMax;
  FLOAT theWeight;  
  
  
  SizeImage(output, n_c, n_r);
  ZeroImage(output);
  
  /* Getting Min Max for input_r signal*/
  if (ymin_r >= ymax_r)
    if( xmin_r >= xmax_r) 
      MinMaxSig(input_r,&xmin_r,&xmax_r,&ymin_r,&ymax_r,&iMin,&iMax,flagCausal);
    else 
      MinMaxSig(input_r,&min,&max,&ymin_r,&ymax_r,&iMin,&iMax,flagCausal);
  else
    if( xmin_r >= xmax_r) 
      MinMaxSig(input_r,&xmin_r,&xmax_r,&min,&max,&iMin,&iMax,flagCausal);
 
  if (ymin_r == ymax_r) {ymin_r -= .5;ymax_r += .5;}


  /* Idem for input_c signal */
  if (ymin_c >= ymax_c)
    if( xmin_c >= xmax_c) 
      MinMaxSig(input_c,&xmin_c,&xmax_c,&ymin_c,&ymax_c,&iMin,&iMax,flagCausal);
    else 
      MinMaxSig(input_c,&min,&max,&ymin_c,&ymax_c,&iMin,&iMax,flagCausal);
  else
    if( xmin_c >= xmax_c) 
      MinMaxSig(input_c,&xmin_c,&xmax_c,&min,&max,&iMin,&iMax,flagCausal);
 
  if (ymin_c == ymax_c) {ymin_c -= .5;ymax_c += .5;}


  /* both input signals must have the same size and when needed side effect are taken from input_r*/
  if(flagCausal == YES) {
    iMin = input_r->firstp;
    iMax = input_r->lastp;
  } else {
    iMin = 0;
    iMax = input_r->size - 1;
  }

  for (i=iMin;i<=iMax;i++) {
    /* are we in the correct x intervals ? */
    if(XSig(input_r,i) > xmax_r || XSig(input_r,i) < xmin_r) 
      continue;
    if(XSig(input_c,i) > xmax_c || XSig(input_c,i) < xmin_c) 
      continue;
    
    /* the indexes */
    index_r = (int)((input_r->Y[i]-ymin_r)*n_r/(ymax_r-ymin_r));
    if (index_r < 0 || index_r > n_r)
      continue;
    if (index_r == n_r)
      index_r = n_r-1;
    
    index_c = (int)((input_c->Y[i]-ymin_c)*n_c/(ymax_c-ymin_c));
    if (index_c < 0 || index_c > n_c)
      continue;
    if (index_c == n_c)
      index_c = n_c-1;
    
    
    output->pixels[n_c*index_r+index_c] += 1.0;
  }
  
}


/* The corresponding command */
void C_Histo2V (char **argv)
{
  SIGNAL input_r,input_c;
  IMAGE output;
  int n_r,n_c,flagCausal=NO;
  FLOAT xmin_r,xmax_r,xmin_c,xmax_c;
  FLOAT ymin_r,ymax_r,ymin_c,ymax_c;
  char opt;
  
  argv = ParseArgv(argv,
		   tSIGNALI, &input_r, tINT, &n_r,
		   tSIGNALI, &input_c, tINT, &n_c,
		   tIMAGE, &output,
		   -1);

  xmin_r = 1;
  xmax_r = -1;
  ymin_r = 1;
  ymax_r = -1;

  xmin_c = 1;
  xmax_c = -1;
  ymin_c = 1;
  ymax_c = -1;

  while(opt = ParseOption(&argv)) { 

    switch(opt) 
      {
      case 'x':
	argv = ParseArgv(argv,
			 tFLOAT,&xmin_r,tFLOAT,&xmax_r,
			 tFLOAT,&xmin_c,tFLOAT,&xmax_c,
			 -1);
	if ( (xmin_r >= xmax_r)
	     || (xmin_c >= xmax_c))
	  Errorf("xmin should be smaller than xmax");
	break;
      case 'y':
	argv = ParseArgv(argv,
			 tFLOAT,&ymin_r,tFLOAT,&ymax_r,
			 tFLOAT,&ymin_c,tFLOAT,&ymax_c,
			 -1);
	if ( (ymin_r >= ymax_r)
	     || (ymin_c >= ymax_c) )
	  Errorf("ymin should be smaller than ymax");
	break;
      case 'c':
	flagCausal=YES;
	break;
      default:
	ErrorOption(opt);
      }
  }
  NoMoreArgs(argv);
  
  if (input_r->size != input_c->size)
    Errorf("Both inputs must have the same size\n");
  
  
  Histo2Sig(output,
	    input_r, n_r, xmin_r, xmax_r, ymin_r, ymax_r,
	    input_c, n_c, xmin_c, xmax_c, ymin_c, ymax_c,
	    flagCausal);
}


static CProc histo2VCommands[] = {
  
  "histo2v",
  C_Histo2V,
  "{{{<signalInRow> <nr> <signalInCol> <nc> <imageOut>  [-x <xMinRow> <xMaxRow> <xMinCol> <xMaxCol>] "
  "[-y <yMinRow> <yMaxRow> <yMinCol> <yMaxCol>] [-c]} {Computes in <imageOut> the <nRow>x<nCol> "
  "branch histogram of the values of <signalIn>s. \n"
  " -x : Only the values between the abscissa <xMin> and <xMax> are taken "
  "into account.\n"
  " -y : Only the values between the ordinate <yMin> and <yMax> are taken "
  "into account. \n"
  "      So the histogram will be made of <n> bars between <ymin> and <ymax>.\n"
  " -c : Only points corresponding to indexes between firstp and lastp (of InRow) are "
  "taken into account.}}}",
	
  NULL,NULL,NULL
};

CProcTable  histo2VTable = {histo2VCommands, "histo2V", "Histogram of 2 variables."};
