/*..........................................................................*/
/*                                                                          */
/*      L a s t W a v e   2 .1                                              */
/*                                                                          */
/*                                                                          */
/*      (C) 2005 Copyright Benjamin Audit                                   */
/*           All Right Reserved.                                            */
/*      Author Benjamin Audit                                               */
/*                                                                          */
/*..........................................................................*/

/*      (C) 1997 Copyright Emmanuel Bacry and Jean-Francois Muzy            */

/****************************************************************************/
/*                                                                          */
/* Reimplementation of the SpDecomp functionality by Emmanuel Bacry and     */
/* Jean-Francois Muzy ((C) 1997)                                            */
/* Decomposition into arbitrary degree increments and there smoothed        */
/* version  e.g. Box, Haar or French Hat function                           */
/*                                                                          */
/****************************************************************************/


#include "lastwave.h"
#include "signals.h"
#include "wtrans1d.h"

extern double round(double x); /* should not be there but was missing in math.h from gcc-4.1 */

/****************************************************************************/
/* The 4 basic convolutions with zero padding border effect.                */
/* Implementation in double, to avoid round-off errors.                     */
/* Scale is noted l and should be < size */

/* increments: I[i+(l+1)//2]-I[i-l//2] */
static void _incr_01(double *in, double *out, int size, int l) {
  int i;
  int k_right = l/2;
  int k_left = (l+1)/2;

  for(i=0;i<k_right;i++){
    out[i] = in[i+k_left];
  }

  for(i=k_right;i<size-k_left;i++){
    out[i] = in[i+k_left]-in[i-k_right];
  }

  for(i=size-k_left;i<size;i++){
    out[i] = -in[i-k_right];
  }

}

/* increments: I[i+l//2]-I[i-(l+1)//2] */
static void _incr_10(double *in, double *out, int size, int l) {
  int i;
  int k_right = (l+1)/2;
  int k_left = l/2;

  for(i=0;i<k_right;i++){
    out[i] = in[i+k_left];
  }

  for(i=k_right;i<size-k_left;i++){
    out[i] = in[i+k_left]-in[i-k_right];
  }

  for(i=size-k_left;i<size;i++){
    out[i] = -in[i-k_right];
  }

}

/* window: One on [(l-1)//2,l//2] */
static void _window_01(double *in, double *out, int size, int l) {
  int i;
  int k_right = l/2;
  int k_left = (l-1)/2;

  out[0] = 0;
  for(i=0;i<=k_left;i++){out[0]+=in[i];}
  for(i=1;i<=k_right;i++){out[i]= out[i-1]+in[i+k_left];}

  for(i=k_right+1;i<size-k_left;i++){
    out[i] = out[i-1]+in[i+k_left]-in[i-k_right-1];
  }
 
  for(i=size-k_left;i<size;i++){
    out[i] = out[i-1]-in[i-k_right-1];
  }
}

/* window: One on [l//2,(l-1)//2] */
static void _window_10(double *in, double *out, int size, int l) {
  int i;
  int k_right = (l-1)/2;
  int k_left = l/2;

  out[0] = 0;
  for(i=0;i<=k_left;i++){out[0]+=in[i];}
  for(i=1;i<=k_right;i++){out[i]= out[i-1]+in[i+k_left];}

  for(i=k_right+1;i<size-k_left;i++){
    out[i] = out[i-1]+in[i+k_left]-in[i-k_right-1];
  }
 
  for(i=size-k_left;i<size;i++){
    out[i] = out[i-1]-in[i-k_right-1];
  }
}

/****************************************************************************/
/* computing border effects size */

static int int_border_left(int l, int incr_degree, int smooth_order) {
  int n,r,m,s;

  /* incr_degree = 2*n+r with r = 0 or 1*/
  n = incr_degree/2;
  r = incr_degree%2;

  /* smooth_order = 2*m+s with s = 0 or 1*/
  m = smooth_order/2;
  s = smooth_order%2;

  /* border_left = sum filter_size_right:
     (n+r) _incr_01, n _incr_10, (m+s)_window_01 and m _window_10 */
  return (l/2)*(n+r)+((l+1)/2)*n+(l/2)*(m+s)+((l-1)/2)*m;
}

static int int_border_right(int l, int incr_degree, int smooth_order) {
  int n,r,m,s;

  /* incr_degree = 2*n+r with r = 0 or 1*/
  n = incr_degree/2;
  r = incr_degree%2;

  /* smooth_order = 2*m+s with s = 0 or 1*/
  m = smooth_order/2;
  s = smooth_order%2;

  /* border_right = sum filter_size_left:
     (n+r) _incr_01, n _incr_10, (m+s)_window_01 and m _window_10 */
  return ((l+1)/2)*(n+r)+(l/2)*n+((l-1)/2)*(m+s)+(l/2)*m;
}

/****************************************************************************/
/* Initialisation of double *input given a border type */
static void init_with_border(SIGNAL in, double *input, int border_type,
			     int border_left, int border_right) {
  int i,j;
  
  /* Middle */
  for(i=border_left;i<in->size+border_left;i++) {
    input[i] = (double) in->Y[i-border_left];
  }
  
  /* Left */
  switch( border_type) {
  case BorderPad :
    for(i=0;i<border_left;i++) {input[i] = (double) in->Y[0];}
    break;
    
  case BorderPad0 :
    for(i=0;i<border_left;i++) {input[i] = 0.0;}
    break;
    
  case BorderPeriodic :
    for(i=0;i<border_left;i++) {
      j = (in->size+i-border_left)%in->size;
      /* Complication due to unclear def of % in C for negative numerator */
      if(j<0) {j+=in->size;}
      input[i] = (double) in->Y[j];
    }
    break;
    
  case BorderMirror :
    for(i=0;i<border_left;i++) {
      j = (border_left-i)%(2*in->size);
      if(j>=in->size) {j = 2*in->size-1-j;}
      input[i] = (double) in->Y[j];
    }
    break;
    
  default:
    Errorf("init_with_border() : Bad Border type !");
  }
  
  /* Right */
  switch( border_type) {
  case BorderPad :
    for(i=0;i<border_right;i++) {input[i+in->size+border_left] = (double) in->Y[in->size-1];}
    break;
    
  case BorderPad0 :
    for(i=0;i<border_right;i++) {input[i+in->size+border_left] = 0.0;}
    break;
    
  case BorderPeriodic :
    for(i=0;i<border_right;i++) {
      j = i%in->size;
      input[i+in->size+border_left] = (double) in->Y[j];
    }
    break;
    
  case BorderMirror :
    for(i=0;i<border_right;i++) {
      j = (in->size-1-i-1)%(2*in->size);
      if(j<0) {j+=2*in->size;} /* Complication due to unclear def of % in C for negative numerator */
      if(j>=in->size) {j = 2*in->size-1-j;}
      input[i+in->size+border_left] = (double) in->Y[j];
    }
    break;
    
  default:
    Errorf("init_with_border() : Bad Border type !");
  }
}

/****************************************************************************/
/* Basic Convolution */

/* l should be in [1,size-1] */
/* WARNING: input will be used as a work space. */
/*          Return value is either input or work_space depending */
/*          on where the results is (avoiding copying).          */
static double *_sp_convol(double *input, double *work_space, int size,int l,int incr_degree,int smooth_order) {
  int k,n,r,m,s;
  double *temp;


  /*********************************/
  /* incr_degree = 2*n+r with r = 0 or 1*/
  n = incr_degree/2;
  r = incr_degree%2;
  /* n+r convol with _incr_01 */
  for(k=0;k<n+r;k++){
    _incr_01(input,work_space,size,l);
    temp = work_space;
    work_space = input;
    input = temp;
  }
  /* n convol with _incr_10 */
  for(k=0;k<n;k++){
    _incr_10(input,work_space,size,l);
    temp = work_space;
    work_space = input;
    input = temp;
  }

  /*********************************/
  /* smooth_order = 2*m+s with s = 0 or 1*/
  m = smooth_order/2;
  s = smooth_order%2;
  /* m+s convol with _window_01 */
  for(k=0;k<m+s;k++){
    _window_01(input,work_space,size,l);
    temp = work_space;
    work_space = input;
    input = temp;
  }
  /* m convol with _window_10 */
  for(k=0;k<m;k++){
    _window_10(input,work_space,size,l);
    temp = work_space;
    work_space = input;
    input = temp;
  }


  return input;
}

/***************************************************************/
/* When smoothing twice or more (smooth_order>=2), there is a  */
/* sampling pb at small scales. To avoid to much bias in the   */
/* scalings we force the L2 (or L1) norm to scale like l.      */
/*                                                             */
/* By default, we do L2 normalisation because it also          */
/* preserves the variance for white noise.                     */
/* Warning: Lp norm of unsmoothed wavelets has to be computed  */
/*          differently than for smoothed wavelets:            */
/*          Unsmoothed case: we want sum psi_l^p = 1           */
/*          Smoothed case: we want l^(-1) sum psi_l^p = 1      */
/***************************************************************/
double sp_l2_norm(int l,int incr_degree,int smooth_order) {
  int i, border_left, border_right, size_tot;
  double l2;
  double *input, *work_space, *result;
  /* border sizes */
  border_left = int_border_left(l,incr_degree,smooth_order);
  border_right = int_border_right(l,incr_degree,smooth_order);
  /* Printf("%d %d\n\n", border_left, border_right); */

  /* inits */
  size_tot = 1+border_left+border_right;
  input = (double *) Malloc(size_tot*sizeof(double));
  work_space = (double *) Malloc(size_tot*sizeof(double));

  for(i=0;i<size_tot;i++) {input[i] = 0.0;}
  input[border_right] = 1.0;

  /* convol */
  result = _sp_convol(input, work_space, size_tot, l, incr_degree, smooth_order);

  /* the l2 norm */
  l2 = 0.0;
  for(i=0;i<size_tot;i++) {
    /* Printf("%d %.4f\n",i,result[i]); */
    l2 += result[i]*result[i];
  }
  if(smooth_order > 0) {l2 /= l;}

  /* Clean up */
  Free(input);
  Free(work_space);

  return sqrt(l2);
}

double sp_l1_norm(int l,int incr_degree,int smooth_order) {
  int i, border_left, border_right, size_tot;
  double l1;
  double *input, *work_space, *result;
  /* border sizes */
  border_left = int_border_left(l,incr_degree,smooth_order);
  border_right = int_border_right(l,incr_degree,smooth_order);
  /* Printf("%d %d\n\n", border_left, border_right); */

  /* inits */
  size_tot = 1+border_left+border_right;
  input = (double *) Malloc(size_tot*sizeof(double));
  work_space = (double *) Malloc(size_tot*sizeof(double));

  for(i=0;i<size_tot;i++) {input[i] = 0.0;}
  input[border_right] = 1.0;

  /* convol */
  result = _sp_convol(input, work_space, size_tot, l, incr_degree, smooth_order);

  /* the l1 norm */
  l1 = 0.0;
  for(i=0;i<size_tot;i++) {
    /* Printf("%d %.4f\n",i,result[i]); */
    l1 += fabs(result[i]);
  }
  if(smooth_order > 0) {l1 /= l;}

  /* Clean up */
  Free(input);
  Free(work_space);

  return l1;
}

/***************************************************************/
/* Computing normalisation so that for wavelet of incr_degree  */
/* i.e. with n = incr_degree null moment, we have :            */
/* lder = \int x^n\psi(x) dx = 1                               */
/* At scale l, we compute lder = \int (x/l)^n\psi_l(x) dx/l    */
/* This normalisation is useful because it allows extracting   */
/* the nth derivative of a signal at scale l:                  */
/*   \int f(x)\psi_{b,a}(x)dx ~ l^{n+1} f^{(n)}(b)/n!          */
/***************************************************************/
double sp_der_norm(int l,int incr_degree,int smooth_order) {
  int i, k, border_left, border_right, size_tot;
  double lder, val;
  double *input, *work_space, *result;
  /* border sizes */
  border_left = int_border_left(l,incr_degree,smooth_order);
  border_right = int_border_right(l,incr_degree,smooth_order);
  /* Printf("%d %d\n\n", border_left, border_right); */

  /* inits */
  size_tot = 1+border_left+border_right;
  input = (double *) Malloc(size_tot*sizeof(double));
  work_space = (double *) Malloc(size_tot*sizeof(double));

  for(i=0;i<size_tot;i++) {
    val = 1.0*(i-border_left)/l;
    input[i] = 1.0/l;
    for(k=0;k<incr_degree;k++) {input[i]*=val;}
  }

  /* convol */
  result = _sp_convol(input, work_space, size_tot, l, incr_degree, smooth_order);

  /* the lder */
  lder = result[border_left];
  /* Printf("\n%f\n", lder); */

  /* if(smooth_order > 0) {l2 /= l;} */

  /* Clean up */
  Free(input);
  Free(work_space);

  return lder;
}

/***************************************************************/
/*  This function is used for computing the convolution at     */
/*  one scale only                                             */
/***************************************************************/

void C_SpConvol(char **argv)
{
  int l, incr_degree, smooth_order;
  int i, k, border_left, border_right, size_tot, border_type, flag_l1_norm, flag_lder_norm;
  char opt;
  char *borderName;
  double *result, *work_space, norm;
  SIGNAL in,out;

  argv = ParseArgv(argv,tSIGNALI,&in,tINT,&l,tINT,&incr_degree,tINT,&smooth_order,-1);
  if(l<1) {Errorf("scale <l> should be >= 1 !");}
  if (l>=in->size) {Errorf("scale <l> should be < than <size signal> !");}
  if(incr_degree<0) {Errorf("increment degree should be >= 0 !");}
  if(smooth_order<0) {Errorf("smooth order should be >= 0 !");}

  border_type = BorderMirror;
  flag_l1_norm = 0;
  flag_lder_norm = 0;
  while(opt = ParseOption(&argv)) {
    switch(opt) {
    case 'b':
      argv = ParseArgv(argv,tSTR,&borderName,-1);
      if(!strcmp(borderName,"pad"))
	border_type =  BorderPad;
      else if(!strcmp(borderName,"mir"))
	border_type = BorderMirror;
      else if(!strcmp(borderName,"pad0"))
	border_type = BorderPad0;
      else if(!strcmp(borderName,"per"))
	border_type = BorderPeriodic;
      else 
	Errorf("Undefined border effect: %s",borderName);
      break;
    case 'l':
      flag_l1_norm = 1;
      break;
    case 'd':
      flag_lder_norm = 1;
      break;
    default: 
      ErrorOption(opt);
    }
  }    
  NoMoreArgs(argv);

  if (flag_lder_norm && flag_l1_norm) {Errorf("C_SpConvol(): -l and -d are not compatible !");}

  /*****/
  /* border sizes */
  border_left = int_border_left(l,incr_degree,smooth_order);
  border_right = int_border_right(l,incr_degree,smooth_order);
  /*****/

  /****/
  /* Inits */
  size_tot = in->size+border_left+border_right;
  result = (double *) TMalloc(size_tot*sizeof(double));
  work_space = (double *) TMalloc(size_tot*sizeof(double));

  init_with_border(in,result,border_type,border_left,border_right);

  /****/
  /* Convol */
  result = _sp_convol(result,work_space,size_tot,l,incr_degree,smooth_order);

  /****/
  /* Set Result */
  out = TNewSignal();
  SizeSignal(out,in->size,YSIG);
  
  if (flag_l1_norm) {
    norm = 1.0/sp_l1_norm(l, incr_degree, smooth_order);
  } else if (flag_lder_norm) {
    norm = l*sp_der_norm(l, incr_degree, smooth_order);
    for(k=0;k<incr_degree;k++) {norm*=l;}
    norm = 1.0/norm;
  } else {
    norm = 1.0/sp_l2_norm(l, incr_degree, smooth_order);
  }
  /*   Printf("%.4f\n",norm); */

  for(i=0;i<in->size;i++) {
    out->Y[i] = (FLOAT) (result[i+border_left]*norm);
  }
  out->firstp = border_left;
  out->lastp = in->size-border_right-1;

  SetResultValue(out);
}

/***************************************************************/
/*  This function is used for computing the wavelet transform  */
/***************************************************************/
void SpDecomp(WTRANS wtrans,FLOAT aMin,int nOct,int nVoi,
	      int incr_degree, int smooth_order,
	      int border_type, FLOAT expo, int flag_lder_norm) {
  int o,v,l;
  int border_left, border_right, i, size;
  double a,mult;
  double *input, *work_space, *result;

  /* Some init */
  wtrans->nOct = nOct;
  wtrans->nVoice = nVoi;
  wtrans->aMin = aMin; 
  wtrans->exponent = expo;
  wtrans->x0 = wtrans->A[0][0]->x0;
  wtrans->dx = wtrans->A[0][0]->dx;
  wtrans->size = wtrans->A[0][0]->size;
  wtrans->type = W_CONT;
 
  /***********/
  /* We allocate memory for the max scale -> One memory allocation */
  a = aMin*pow(2.,nOct);
  l = round(a);
  if(l<1) l=1;
  size = wtrans->A[0][0]->size
    +int_border_left(l,incr_degree,smooth_order)
    +int_border_right(l,incr_degree,smooth_order);
  input = (double *) TMalloc(size*sizeof(double));
  work_space = (double *) TMalloc(size*sizeof(double));
  /***********/

  size = wtrans->A[0][0]->size;

  for (o = 1; o <= nOct ; o++) {
    for(v=0; v<nVoi; v++) {
      if (v == 0 && o == 1) Printf("Start octave %d",o);
      else if (v == 0 && o != 1) Printf("\nStart octave %d",o);
      else Printf(".");
      Flush();

      /*****/
      /* The scale */
      a = aMin*pow(2.,(o-1.)+v/(1.0*nVoi));
      l = round(a);
      if(l<1) l=1;
      /* The borders */
      border_left = int_border_left(l,incr_degree,smooth_order);
      border_right = int_border_right(l,incr_degree,smooth_order);
      /* Init input */
      init_with_border(wtrans->A[0][0], input, border_type, border_left, border_right);
      /* The convol */
      result = _sp_convol(input, work_space, size+border_left+border_right,
			  l, incr_degree, smooth_order);

      /* Set Result */
      SizeSignal(wtrans->D[o][v],size,YSIG);

      wtrans->D[o][v]->dx  = wtrans->A[0][0]->dx;
      wtrans->D[o][v]->x0  = wtrans->A[0][0]->x0;
      
      wtrans->D[o][v]->firstp  = wtrans->A[0][0]->firstp+border_left;
      wtrans->D[o][v]->lastp  = wtrans->A[0][0]->lastp-border_right;

      if (flag_lder_norm) {
	mult = pow(1.0*l, expo)/sp_der_norm(l, incr_degree, smooth_order);
      } else {
	mult = pow(1.0*l, expo)/sp_l2_norm(l, incr_degree, smooth_order);
      }
      for(i=0;i<size;i++) {
	wtrans->D[o][v]->Y[i] = (FLOAT) (result[i+border_left]*mult);
      }        
    } 
  }
  Printf("\n");
  
}

/* The corresponding command */
void C_SpDecomp(char **argv)
{
  int nOct,nVoi;
  int incr_degree,smooth_order;
  int border_type, flag_lder_norm;
  char *borderName;
  char opt;
  FLOAT aMin,expo,time;
  WTRANS wtrans;
  char tempWNname[20];

  argv = ParseArgv(argv,tWTRANS_,NULL,&wtrans,
		   tFLOAT,&aMin,tINT,&nOct,tINT,&nVoi,
		   tINT,&incr_degree,tINT,&smooth_order,-1);
  
  if (wtrans == NULL) wtrans = GetWtransCur();

  if (wtrans->A[0][0]->size == 0) Errorf("No signal to analyze");
  
  if (!INRANGE(1,nOct,NOCT-1)) Errorf("Bad range for <noct>");
  if (!INRANGE(1,nVoi,NVOICE)) Errorf("Bad range for <nvoice> \n");
  
  if (aMin<1.0)
    Warningf("You are using a value for aMin < 1.0 !");
  if ( ((int) round(aMin*pow(2.0,nOct - 1.0/((double) nVoi)))) > wtrans->A[0][0]->size) 
    Warningf("The largest scale is bigger than the signal size !");

  /***************/
  /* The Options */
  /***************/
  expo = -1.;
  border_type = BorderMirror;
  flag_lder_norm = 0;
  while(opt = ParseOption(&argv)) {
    switch(opt) {
    case 'e':
      argv = ParseArgv(argv,tFLOAT,&expo,-1);
      break;
    case 'b':
      argv = ParseArgv(argv,tSTR,&borderName,-1);
      if(!strcmp(borderName,"pad"))
	border_type =  BorderPad;
      else if(!strcmp(borderName,"mir"))
	border_type = BorderMirror;
      else if(!strcmp(borderName,"pad0"))
	border_type = BorderPad0;
      else if(!strcmp(borderName,"per"))
	border_type = BorderPeriodic;
      else 
	Errorf("Undefined border effect: %s",borderName);
      break;
    case 'd':
      flag_lder_norm = 1;
      break;
    default: 
      ErrorOption(opt);
    }
  }    
  NoMoreArgs(argv);
  
  
  if (wtrans->wName != NULL) Free(wtrans->wName);
  snprintf(tempWNname,sizeof(tempWNname),"sp_d%d_o%d", incr_degree, smooth_order);
  wtrans->wName = CopyStr(tempWNname);
  
  time = MyTime();

  SpDecomp(wtrans, aMin, nOct, nVoi, incr_degree, smooth_order, border_type, expo, flag_lder_norm);

  SetResultFloat(MyTime()-time);                 
}

/*****************************************************************************************/
static CProc spCommands[] = {

  "spconvol",C_SpConvol,
  "{{{<sig_in> <scale> <incr. degree> <smooth order> [-b <border type>='mir'] [-l|-d]} "
  "{Return the increments of degree <degree> at scale <l> smoothed <order> times. "
  "By default, L2 norm is forced to <scale>. Use -l to force L1 norm or -d to for 'derivative norm'.\n"
  "   Border type can be on of:\n"
  "      'mir': mirror\n"
  "      'per': periodic\n"
  "      'pad': padding\n"
  "      'pad0': padding with 0\n"
  "}}}",

  "spdecomp",C_SpDecomp,
  "{{{[<wtrans>] <aMin> <noct> <nvoice> <incr. degree> <smooth order> [-b <border type>='mir'] [-e <expo>=-1] } "
  "{Computes the wavelet decomposition with analysing wavelet the increment of degree <degree> smoothed <order> times."
  "By default, L2 norm is forced to be constant across scales. Use -d to force the 'derivative norm'.\n"
  "  -e: normalisation exponent.\n"
  "  -b: border effects ('per'=periodic, 'mir'=mirror, 'pad0'=padding with 0 values, 'pad'=padding" 
  " with constant extremity values so that the signal is continuous)."
  "}}}",

  NULL,NULL,NULL
};

CProcTable spTable = 
{spCommands, "Sp",
 "SpDecomposition commands."};

